/**
  *  \file client/chart/standardmode.cc
  *  \brief Starchart Mode: WStandardChartMode
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/chart/distancemode.h"
#include "client/chart/markermode.h"
#include "client/chart/messagemode.h"
#include "client/chart/selectmode.h"
#include "client/chart/standardmode.h"
#include "client/chartloc.h"
#include "client/dialogs/chartconfig.h"
#include "client/dialogs/gotoxydlg.h"
#include "client/dialogs/labeldlg.h"
#include "client/dialogs/tagcreate.h"
#include "client/dialogs/tagselect.h"
#include "client/dlg-draw.h"
#include "client/markerconfig.h"
#include "client/marks.h"
#include "client/screens.h"
#include "client/screens/ionscreen.h"
#include "client/screens/minescreen.h"
#include "client/screens/ufoscreen.h"
#include "client/widgets/colorselect.h"
#include "client/widgets/markertype.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/lock.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "gfx/update.h"
#include "int/if/userif.h"
#include "int/simple.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dialogs/rangeedit.h"
#include "game/pref.h"

/******************************* Functions *******************************/

/* Remap vk_WheelDown, vk_WheelUp according to user configuration. */
static uint32_t
remapWheelKey(uint32_t key)
{
    uint32_t ctrlMod = (key & ss_Ctrl);
    switch (key & ~ss_Ctrl) {
     case vk_WheelDown:
        switch (getUserPreferences().ChartWheel()) {
         case GUserPreferences::cw_Zoom:  return ctrlMod + '-';
         case GUserPreferences::cw_Tab:   return ctrlMod + SDLK_TAB;
         case GUserPreferences::cw_Page:  return ctrlMod + SDLK_PAGEDOWN;
        }
        break;

     case vk_WheelUp:
        switch (getUserPreferences().ChartWheel()) {
         case GUserPreferences::cw_Zoom:  return ctrlMod + '+';
         case GUserPreferences::cw_Tab:   return ctrlMod + ss_Shift + SDLK_TAB;
         case GUserPreferences::cw_Page:  return ctrlMod + SDLK_PAGEUP;
        }
        break;

     default:
        break;
    }
    return key;
}


static void
drawCursor(GfxCanvas& can, GfxPoint pt, int angle)
{
    static const int SIZE = 15;
    static const double FACTOR = 5 * PI / 180;

    angle = -angle;

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.setColor(COLOR_RED);

    /* it turns out that scaling the cursor doesn't look that good. */
    const double scaled_size = SIZE;

    for (int i = 0; i < 2; ++i) {
        double d = angle * PI / 180;
        drawLine(ctx,
                 GfxPoint(pt.x + roundToInt(4*std::sin(d)), pt.y + roundToInt(4*std::cos(d))),
                 GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d)), pt.y + roundToInt(scaled_size*std::cos(d))));
        for (int j = -8; j < 8; ++j) {
            drawLine(ctx,
                     GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d-FACTOR*j)),
                              pt.y + roundToInt(scaled_size*std::cos(d-FACTOR*j))),
                     GfxPoint(pt.x + roundToInt(scaled_size*std::sin(d-FACTOR*j+FACTOR)),
                              pt.y + roundToInt(scaled_size*std::cos(d-FACTOR*j+FACTOR))));
        }
        angle += 180;
    }
}

static void
drawObjectList(GfxCanvas& can, GChartLocation& loc)
{
    const GObjectList& chart_objlist = loc.getObjectList();

    if (!chart_objlist.size())
        return;

    int lines = loc.getObjectListHeight();
    int top   = loc.getObjectListTop();

    GfxPoint pt = loc.getViewport().getCenter();
    int h = font_heights[FONT_NORMAL]*lines;
    int y = pt.y - h/2;

    WStarchartTileSkin::drawFancyBar(can, GfxRect(pt.x - 300, y-2, 280, h+4));

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = 0; i < lines; ++i) {
        GObjectReference id = chart_objlist[i + top - 1];
        UIColor::Color color = (id.isGameObject()
                                ? getPlayerColor(id->getOwner())
                                : UIColor::tc_Static);
        if (id == loc.getCurrentObject()) {
            if (ui_screen_bits <= 8)
                can.drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                            standard_colors[COLOR_SHIELD+4], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
            else
                can.drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                            standard_colors[COLOR_SHIELD+8], 0, SOLID_FILL, 128);
        }

        int wi = 276;
        if (id.isGameObject() && id->isMarked()) {
            ctx.setColor(COLOR_YELLOW);
            drawSelection(ctx, GfxPoint(pt.x-wi-17, y+i*font_heights[FONT_NORMAL]+7), 1, 2);
            wi -= 15;
        }
        ctx.setColor(BLACK_COLOR_SET[color]);
        outTextF(ctx, pt.x-wi-22, y+i*font_heights[FONT_NORMAL], wi, id->getName(GObject::LongName));
    }
}

/** Page through objects.
    \param loc Chart location object to operate on
    \param delta Browse direction (+1/-1)
    \param marked_only true to accept only marked objects
    \param any true to accept any race's object. Default is to accept only played objects
           when starting at a played object.

    FIXME: PCC 1.x behaves a little different. With any=false, PCC 1.x would only accept
    played objects no matter where we start, but it would "degrade" to accept non-played
    objects when there are no played ones. */
static void
doPage(GChartLocation& loc, int delta, bool marked_only, bool any)
{
    GObjectType* type;

    /* Fetch current object, and associated type object */
    GObjectReference ref = loc.getCurrentObject();
    if (!ref.isValid() || !ref.isGameObject())
        return;

    if (GShip* sh = dynamic_cast<GShip*>(&*ref)) {
        // Iterate through all ships or player ships
        if (sh->isPlayable(GObject::ReadOnly) && !any)
            type = &current_ship_type;
        else
            type = &current_anyship_type;
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&*ref)) {
        // Iterate through all planets or player planets
        if (pl->isPlayable(GObject::ReadOnly) && !any)
            type = &current_planet_type;
        else
            type = &current_anyplanet_type;
    } else {
        // Iterate through whatever it is
        type = &ref.getObjectType();
    }

    /* Get next object */
    int next;
    if (delta > 0) {
        next = type->findNextIndexWrap(ref.getObjectIndex(), marked_only);
    } else {
        next = type->findPreviousIndexWrap(ref.getObjectIndex(), marked_only);
    }

    /* Valid? */
    if (type->isValidIndex(next))
        loc.setCurrentObject(type->getObjectByIndex(next));
}

static void
doChartLock(GChartLocation& loc, TLockMode lm, bool marked)
{
    GPoint pt = loc.getLocation();
    GLockData locker(pt, marked);
    findItem(getDisplayedTurn().getCurrentUniverse(), locker, lm);
    if (const GMapObject* obj = locker.getObject()) {
        loc.setCurrentObject(*obj);
    }
    loc.setLocation(locker.getPoint().getSimpleNearestAlias(loc.getLocation()));
}

static void
doGoTo(GChartLocation& loc)
{
    GPoint pt;
    if (doGoToXYDialog(pt))
        loc.setLocation(pt);
}

static void
doColorizeAdjacent(GDrawingContainer& c, GPoint pt, uint8 color)
{
    GDrawingContainer::iterator i = c.begin(), e = c.end();
    pt = pt.getSimpleCanonicalLocation();
    while (i != e) {
        GDrawing& d = **i;
        if (d.isVisible() && d.getType() == LineDrawing && d.getColor() != color) {
            if (d.getPos().getSimpleCanonicalLocation() == pt) {
                // Starts at given position
                d.setColor(color);
                pt = d.getPos2().getSimpleCanonicalLocation();
                i = c.begin();
            } else if (d.getPos2().getSimpleCanonicalLocation() == pt) {
                // Ends at given position
                d.setColor(color);
                pt = d.getPos().getSimpleCanonicalLocation();
                i = c.begin();
            } else {
                // No match
                ++i;
            }
        } else {
            ++i;
        }
    }
}

static void
doTagAdjacent(GDrawingContainer& c, GPoint pt, atom_t tag)
{
    GDrawingContainer::iterator i = c.begin(), e = c.end();
    pt = pt.getSimpleCanonicalLocation();
    while (i != e) {
        GDrawing& d = **i;
        if (d.isVisible() && d.getType() == LineDrawing && d.getTag() != tag) {
            if (d.getPos().getSimpleCanonicalLocation() == pt) {
                // Starts at given position
                d.setTag(tag);
                pt = d.getPos2().getSimpleCanonicalLocation();
                i = c.begin();
            } else if (d.getPos2().getSimpleCanonicalLocation() == pt) {
                // Ends at given position
                d.setTag(tag);
                pt = d.getPos().getSimpleCanonicalLocation();
                i = c.begin();
            } else {
                // No match
                ++i;
            }
        } else {
            ++i;
        }
    }
}

static void
doColorChange(GChartLocation& loc)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getSimpleCanonicalLocation());
    if (i == c.end())
        return;

    // Change color
    WColorSelector csel;
    csel.setColor(GDrawing::getUserColorFromColor((*i)->getColor()));
    int n = csel.doStandardDialog(_("Drawing Color"), ((*i)->getType() == LineDrawing));
    if (n == csel.rCancel)
        return;

    uint8 new_color = GDrawing::getColorFromUserColor(csel.getColor());
    (*i)->setColor(new_color);
    if (n == csel.rAdjacent) {
        doColorizeAdjacent(c, (*i)->getPos(), new_color);
        doColorizeAdjacent(c, (*i)->getPos2(), new_color);
    }

    loc.markChanged();
}

static void
doTagChange(GChartLocation& loc)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getSimpleCanonicalLocation());
    if (i == c.end())
        return;

    // Change tag
    atom_t tag = (*i)->getTag();
    bool adj = false;
    bool* pAdjacent = (*i)->getType() == LineDrawing ? &adj : 0;
    if (createMarkerTag(_("Drawing Tag"), tag, pAdjacent)) {
        (*i)->setTag(tag);
        if (adj) {
            doTagAdjacent(c, (*i)->getPos(), tag);
            doTagAdjacent(c, (*i)->getPos2(), tag);
        }
        loc.markChanged();
    }
}

static void
doMoveMarker(GChartLocation& loc, WChartModeList& modes)
{
    // Not possible if drawings not visible
    GChartViewport& viewport = loc.getViewport();
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find drawing here
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.begin();
    while (i != c.end() && ((*i)->getPos().getSimpleCanonicalLocation() != loc.getLocation().getSimpleCanonicalLocation() || (*i)->getType() != MarkerDrawing))
        ++i;

    if (i != c.end())
        WMarkerChartMode::moveMarker(modes, i);
}

static void
doDraw(GChartLocation& loc, WChartModeList& modes)
{
    // Ask user what to do
    WNewDrawingDialog ndd;
    ndd.setColor(9);
    if (!ndd.runDialog())
        return;

    // Draw it
    GDrawing d(loc.getLocation().getSimpleCanonicalLocation(), ndd.getType());
    d.setColor(GDrawing::getColorFromUserColor(ndd.getColor()));
    d.setTag(ndd.getTag());
    switch (ndd.getType()) {
     case LineDrawing:
     case RectangleDrawing:
        d.setPos2(d.getPos());
        WMarkerChartMode::drawLineRectangle(modes, getDisplayedTurn().getCurrentUniverse().getDrawings().add(d));
        break;
     case MarkerDrawing: {
        WMarkerTypeSelector mts;
        WColorSelector cs;
        if (mts.doStandardDialog(_("New marker"))) {
            d.setMarkerKind(mts.getMarkerType());
            cs.setColor(GDrawing::getUserColorFromColor(d.getColor()));
            if (cs.doStandardDialog(_("New marker"), false) == cs.rOK) {
                d.setColor(GDrawing::getColorFromUserColor(cs.getColor()));
                getDisplayedTurn().getCurrentUniverse().getDrawings().add(d);
                loc.markChanged();
            }
        }
        delete modes.getModeByKind(ChartMode_PrimaryMode);
        break;
     }
     case CircleDrawing:
        d.setCircleRadius(10);
        WMarkerChartMode::drawCircle(modes, getDisplayedTurn().getCurrentUniverse().getDrawings().add(d));
        break;
    }
}

static void
doDistance(GChartLocation& loc, WChartModeList& modes)
{
    modes.pushMode(*new WDistanceChartMode(loc.getLocation(), loc.getCurrentShipId()));
    if (loc.getCurrentShipId() != 0) {
        GShip& sh = getDisplayedTurn().getCurrentUniverse().getShip(loc.getCurrentShipId());
        if (sh.getShipKind() == GShip::CurrentShip && sh.getPos() == loc.getLocation()) {
            loc.setLocation(sh.getWaypoint());
        }
    }
}

static void
doMarkRange(GChartLocation& loc, WChartModeList& modes)
{
    modes.pushMode(*new WSelectChartMode(loc.getLocation()));
}

static void
doDelete(GChartLocation& loc, WChartModeList& modes)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getSimpleCanonicalLocation());
    if (i == c.end())
        return;

    WMarkerChartMode::deleteDrawing(modes, i);
}

static void
tryEditMarkerComment(GChartLocation& loc)
{
    /* Find a marker */
    GPoint pt = loc.getLocation().getSimpleCanonicalLocation();
    GDrawingContainer& drawings = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator p = drawings.begin();
    while (p != drawings.end()
           && !((*p)->isVisible()
                && (*p)->getPos().getSimpleCanonicalLocation() == pt
                && (*p)->getType() == MarkerDrawing))
        ++p;

    /* Edit it */
    if (p != drawings.end()) {
        editMarkerComment(**p);
        loc.markChanged();
    }
}

static void
createCannedMarker(int id, GChartLocation& loc)
{
    ConfigMarkerOption& opt = ConfigMarkerOption::getCannedMarker(id % ConfigMarkerOption::NUM_CANNED_MARKERS);

    // Create marker
    GDrawing marker(loc.getLocation().getSimpleCanonicalLocation(), MarkerDrawing);
    marker.setMarkerKind(opt.getType());
    marker.setColor(marker.getColorFromUserColor(opt.getColorIndex()));
    getDisplayedTurn().getCurrentUniverse().getDrawings().add(marker);

    // Redraw
    loc.markChanged();
}

static GUfo*
findUfoOtherEnd(GPoint pt)
{
    const GUfoType& ufos = getDisplayedTurn().getCurrentUniverse().ty_ufos;
    for (GUfoType::const_iterator it = ufos.begin(), end = ufos.end(); it != end; ++it) {
        if (distanceSquared(it->getPos(), pt) <= it->getRadiusSquared()) {
            if (GUfo* otherEnd = it->getOtherEnd()) {
                return otherEnd;
            }
        }
    }
    return 0;
}

/*************************** WStandardChartMode **************************/

/** Constructor. */
WStandardChartMode::WStandardChartMode()
    : WChartMode(ChartMode_BaseMode),
      cursor_angle(0),
      overlays(0)
{ }

/** Destructor. */
WStandardChartMode::~WStandardChartMode()
{
    delete overlays;
}

void
WStandardChartMode::onMove(GChartLocation& loc)
{
    // Build overlay tile
    delete overlays;
    overlays = 0;
    if (loc.getViewName().size() != 0) {
        overlays = new WStarchartTileContainer();
        createTileLayout(*overlays, loc.getViewName(), const_cast<GObjectSelection&>(loc.getObjectSelection()));
        overlays->doLayout(loc.getViewport());
    }

    // Do we have an overlay?
    if (overlays != 0) {
        if (overlays->getFirstChild() == 0) {
            delete overlays;
            overlays = 0;
        }
    }

    // Call next mode
    if (WChartMode* cm = getNextMode())
        cm->onMove(loc);
}


void
WStandardChartMode::drawBelow(GfxCanvas& can, GChartLocation& loc)
{
    // Draw ranges
    if (!ranges.isEmpty()) {
        GChartViewport& vp = loc.getViewport();
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.setColor(COLOR_SHIELD + 4);
        ctx.setFillPattern(GRAY50_FILL);

        // Make sure the fill pattern moves with scrolling.
        // This places the GRAY50_FILL just opposite to the web mine field filling
        // (see GChartViewport::drawMines).
        GfxPoint adj(vp.scaleXY(GPoint(0, 0)));
        if (((adj.x + adj.y) & 1) == 0) {
            ctx.getFillPattern().shiftUp(1);
        }

        drawBar(ctx, vp.view);

        ctx.setFillPattern(SOLID_FILL);
        ctx.setColor(0);

        for (int img = 0, nimg = chart_conf.getNumRectangularImages(); img < nimg; ++img) {
            // Check whether this map image is visible
            GfxRect a;
            a.include(vp.scaleXY(chart_conf.getSimplePointAlias(ranges.getMinimum(), img)));
            a.include(vp.scaleXY(chart_conf.getSimplePointAlias(ranges.getMaximum(), img)));
            a.intersect(vp.view);
            if (a.exists()) {
                // Yes, exists! Draw ranges.
                for (GRangeSet::Iterator_t it = ranges.begin(), e = ranges.end(); it != e; ++it) {
                    drawFilledCircle(ctx,
                                     vp.scaleXY(chart_conf.getSimplePointAlias(it->first, img)),
                                     vp.scale(it->second));

                }
            }
        }
    }

    // Call next mode
    if (WChartMode* cm = getNextMode())
        cm->drawBelow(can, loc);
}

void
WStandardChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    GfxPoint sc = loc.getViewport().scaleXY(loc.getLocation().getSimpleNearestAlias(loc.getCenter()));
    if (!loc.getCurrentObject().isValid()) {
        GfxContext ctx;
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useCanvas(can);
        ctx.setColor(COLOR_BLUE);
        drawHLine(ctx, sc.x-30, sc.y, sc.x-6);
        drawHLine(ctx, sc.x+30, sc.y, sc.x+6);
        drawVLine(ctx, sc.x, sc.y-30, sc.y-6);
        drawVLine(ctx, sc.x, sc.y+30, sc.y+6);
    } else {
        ::drawCursor(can, sc, cursor_angle);
    }
}

void
WStandardChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    drawObjectList(can, loc);

    if (overlays) {
        GfxUpdateOptimizer o(can);
        GfxRect r;
        overlays->drawEverything(o);
        o.clearUpdateRegion(r); /* we forget the region, because the whole screen is refreshed afterwards */
    }

    // Coordinates
    GPoint pt = loc.getLocation().getSimpleCanonicalLocation();
    GPoint pt1 = pt.getCanonicalLocation();
    GfxRect view(loc.getViewport().view);

    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_FIXED]);
    ctx.setColor(COLOR_WHITE);
    ctx.setTextAlign(2, 2);
    if (pt == pt1) {
        outText(ctx, view.getRightX(), view.getBottomY(), format("%4d,%4d", pt.x, pt.y));
    } else {
        outText(ctx, view.getRightX(), view.getBottomY() - font_heights[FONT_NORMAL], format("%4d,%4d", pt.x, pt.y));
        outText(ctx, view.getRightX(), view.getBottomY(), format(_("wraps to %4d,%4d"), pt1.x, pt1.y));
    }

    // Sector Number
    ctx.useFont(*fonts[FONT_NORMAL]);
    int secn = pt.getSectorNumber();
    if (secn != 0) {
        ctx.setTextAlign(0, 2);
        outText(ctx, view.x, view.getBottomY(), format(_("Sector %d"), secn));
    }

    // Filter
    atom_t f = loc.getViewport().drawing_filter;
    if (f != GChartViewport::NO_FILTER) {
        ctx.setTextAlign(0, 0);
        outText(ctx, view.x, view.y, format(_("Drawing filter: showing only %s"),
                                            isAtom(f) ? atomStr(f) : itoa(f)));
    }
}

bool
WStandardChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Tick) {
        if (loc.getCurrentObject().isValid()) {
            /* Advance rotation, and trigger redraw. This used to draw
               in-place; the new way supports real interaction with
               other overlays. */
            cursor_angle = (cursor_angle + 3) % 360;
            GfxPoint sc = loc.getViewport().scaleXY(loc.getLocation());
            getParent()->markDirty(GfxRect(sc.x - 15, sc.y - 15, 31, 31));
        }
    } else if (event.type == ev_Key) {
        // Check keymap
        std::auto_ptr<IntValue> val(IntUserInterfaceBinding::get().getProperty(iuiKeymap));
        IntKeymapValue* kv = dynamic_cast<IntKeymapValue*>(val.get());

        // Do we have a key?
        uint32_t cmd;
        if (kv && isAtom(cmd = kv->getKeymap()->lookupCommand(event.key.code))) {
            executeStatement(atomStr(cmd), event.key.arg);
            return true;
        } else {
            switch (remapWheelKey(event.key.code)) {
             case SDLK_RETURN:
                doChartLock(loc, lm_Right, false);
                return true;
             case SDLK_SPACE:
                doChartLock(loc, lm_Left, false);
                return true;
             case ss_Ctrl + SDLK_RETURN:
                doChartLock(loc, lm_Right, true);
                return true;
             case ss_Ctrl + SDLK_SPACE:
                doChartLock(loc, lm_Left, true);
                return true;
             case '-':
                loc.doZoom(false);
                return true;
             case '+':
                loc.doZoom(true);
                return true;
             case SDLK_TAB:
                loc.doScroll(+1, false);
                return true;
             case ss_Shift + SDLK_TAB:
                loc.doScroll(-1, false);
                return true;
             case ss_Ctrl + SDLK_TAB:
                loc.doScroll(+1, true);
                return true;
             case ss_Shift + ss_Ctrl + SDLK_TAB:
                loc.doScroll(-1, true);
                return true;
             case SDLK_PAGEDOWN:
                doPage(loc, +1, false, false);
                return true;
             case SDLK_PAGEDOWN + ss_Ctrl:
                doPage(loc, +1, true, false);
                return true;
             case SDLK_PAGEUP:
                doPage(loc, -1, false, false);
                return true;
             case SDLK_PAGEUP + ss_Ctrl:
                doPage(loc, -1, true, false);
                return true;
             case SDLK_PAGEDOWN + ss_Shift:
                doPage(loc, +1, false, true);
                return true;
             case SDLK_PAGEDOWN + ss_Ctrl + ss_Shift:
                doPage(loc, +1, true, true);
                return true;
             case SDLK_PAGEUP + ss_Shift:
                doPage(loc, -1, false, true);
                return true;
             case SDLK_PAGEUP + ss_Ctrl + ss_Shift:
                doPage(loc, -1, true, true);
                return true;
             case SDLK_F9:
             case SDLK_F9 + ss_Alt:
                // ship and planet comments are handled through the script interface
                tryEditMarkerComment(loc);
                return true;
             case 'm' + ss_Ctrl:
                createCannedMarker(event.key.arg, loc);
                return true;
             case 'm':
                if (event.key.arg != 0) {
                    createCannedMarker(event.key.arg, loc);
                } else if (int mf = current_minefield_type.findNearestIndex(loc.getLocation())) {
                    current_minefield_selection.setCurrentIndex(mf);
                    doMinefieldScreen();
                } else {
                    WMessageChartMode::showMessage(*getParent(), _("No Minefields"));
                }
                return true;
             case 'u':
                if (int ufo = current_ufo_type.findNearestIndex(loc.getLocation())) {
                    current_ufo_selection.setCurrentIndex(ufo);
                    doUfoScreen();
                } else {
                    WMessageChartMode::showMessage(*getParent(), _("No Ufos"));
                }
                return true;
             case 'i':
                if (int ion = current_ion_type.findNearestIndex(loc.getLocation())) {
                    current_ion_selection.setCurrentIndex(ion);
                    doIonStormScreen();
                } else {
                    WMessageChartMode::showMessage(*getParent(), _("No Ion Storms"));
                }
                return true;
             case 'c':
                doColorChange(loc);
                return true;
             case 't':
                doTagChange(loc);
                return true;
             case 'v':
                doMoveMarker(loc, *getParent());
                return true;
             case 'p':
                doDraw(loc, *getParent());
                return true;
             case 'd':
                doDistance(loc, *getParent());
                return true;
             case 'r':
                doMarkRange(loc, *getParent());
                return true;
             case ss_Ctrl + 'r':
                if (ranges.isEmpty()) {
                    doEditRanges(ranges);
                } else {
                    ranges.clear();
                }
                loc.markChanged();
                return true;
             case 'g':
                doGoTo(loc);
                return true;
             case SDLK_DELETE:
                doDelete(loc, *getParent());
                return true;
             case ss_Alt + 'l':
                editLabelExpressions();
                loc.markChanged();     // Force redraw
                return true;
             case ss_Alt + 'o':
                doChartConfig();
                loc.markChanged();     // Force redraw
                return true;
             case ss_Alt + 'f':
                if (loc.getViewport().drawing_filter == GChartViewport::NO_FILTER) {
                    selectMarkerTag(loc.getViewport().drawing_filter);
                    loc.markChanged();
                } else {
                    loc.getViewport().drawing_filter = GChartViewport::NO_FILTER;
                    loc.markChanged();
                }
                return true;
             case 'x':
                if (GUfo* ufo = findUfoOtherEnd(loc.getLocation())) {
                    loc.setLocation(ufo->getPos());
                } else if (chart_conf.getMode() == GChartConfiguration::Circular) {
                    GPoint pt = chart_conf.getCanonicalLocation(loc.getLocation());
                    GPoint pt2;
                    if (pt != loc.getLocation()) {
                        loc.setLocation(pt);
                    } else if (chart_conf.getPointAlias(pt, pt2, 1, true)) {
                        loc.setLocation(pt2);
                    }
                }
                return true;
             default:
                if ((event.key.code & ss_Alt) != 0) {
                    const GChartViewport& vp = loc.getViewport();
                    if (getChartOpts(false, vp.mult, vp.divi).toggleOptionKey(event.key.code & ~(ss_Alt | ss_Ctrl))) {
                        loc.markChanged();
                        return true;
                    }
                }
                break;
            }
        }
    }
    return false;
}
