/**
  *  \file client/dialogs/tagcreate.cc
  *  \brief Create a Marker Tag
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/tagcreate.h"
#include "client/dialogs/tagselect.h"
#include "client/widgets/helpwidget.h"
#include "game/global.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/statictext.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/widgets/quit.h"

namespace {
    class WCreateMarkerDialog : public UIWindow {
     public:
        WCreateMarkerDialog(const string_t& title,
                            atom_t& out,
                            bool* pAdjacent,
                            StringList& list);
        ~WCreateMarkerDialog();

        atom_t getResult();

     private:
        atom_t& out;
        bool* pAdjacent;
        StringList& list;
        ObjectHolder h;

        UIInputLine* input;
        UIStandardListbox* box;

        void init(StringList& list);

        void onEdit();
        void onMove();
    };

    enum {
        Cancel,
        OK,
        Adjacent
    };
}

inline
WCreateMarkerDialog::WCreateMarkerDialog(const string_t& title,
                                         atom_t& out,
                                         bool* pAdjacent,
                                         StringList& list)
    : UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY),
      out(out),
      pAdjacent(pAdjacent),
      list(list),
      h(),
      input(0),
      box(0)
{
    init(list);
}

WCreateMarkerDialog::~WCreateMarkerDialog()
{ }

inline atom_t
WCreateMarkerDialog::getResult()
{
    int result;
    if (stringValue(input->getText(), result)) {
        return result;
    } else {
        return atom(input->getText());
    }
}

void
WCreateMarkerDialog::init(StringList& list)
{
    // VBox
    //   HBox
    //     UIStaticText "Tag:"
    //     UIInputLine
    //   UIFrameGroup/HBox
    //     UIStandardListbox
    //     UIScrollbar
    //   HBox
    //     "OK"
    //     "Cancel"
    //     "Adj."
    //     UISpacer
    //     "Help"

    setLayout(UIVBoxLayout::instance5);

    input = &h.add(new UIInputLine(200, 20, 0, 0, 1000));
    input->sig_changed.add(this, &WCreateMarkerDialog::onEdit);
    box = &h.add(new UIStandardListbox(1001, 20, 30*font_ems[FONT_NORMAL], list));
    box->sig_moved.add(this, &WCreateMarkerDialog::onMove);

    UIGroup& group1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group1.add(h.add(new UIStaticText(_("Tag:"), UIStaticText::StaticTitle)));
    group1.add(*input);
    add(group1);

    UIFrameGroup& group2 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    group2.add(*box);
    group2.add(h.add(new UIScrollbar(*box)));
    add(group2);

    UIGroup& group3 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group3.add(h.add(new UIButton(OK, _("OK"), SDLK_RETURN)).addStop());
    group3.add(h.add(new UIButton(Cancel, _("Cancel"), SDLK_ESCAPE)).addStop());
    if (pAdjacent) {
        group3.add(h.add(new UIButton(Adjacent, _("Alt-A - Adjacent"), ss_Alt + 'a')).addStop());
    }
    group3.add(h.add(new UISpacer()));
    group3.add(h.add(new UIButton(_("Help"), 'h')));
    add(group3);
    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab)));
    add(h.add(new WHelpWidget("pcc2:drawtag")));
    add(h.add(new UIQuit(Cancel)));

    pack();

    // Set text. This will trigger onEdit and therefore listbox cursor placement.
    input->setText(isAtom(out) ? atomStr(out) : itoa(out));
}

void
WCreateMarkerDialog::onEdit()
{
    string_t text = input->getText();
    int i = 0;
    for (StringList::iterator it = list.begin(), e = list.end(); it != e; ++it, ++i) {
        if (it->second == text) {
            if (i != box->getCurrent()) {
                box->scrollTo(i, 1);
            }
            break;
        }
    }
}

void
WCreateMarkerDialog::onMove()
{
    string_t text = list.get(box->getCurrent());
    if (text != input->getText()) {
        input->setText(text);
    }
}

/** Create a new marker tag.
    \param out [in/out] Tag
    \param pAdjacent [out] If non-null, dialog offers "adjacent";
           true is returned here if that was chosen
    \return true on ok, otherwise false */
bool
createMarkerTag(string_t title, atom_t& out, bool* pAdjacent)
{
    // Need a turn
    if (!haveDisplayedTurn()) {
        return false;
    }

    // Build the list
    StringList list;
    buildCurrentTagList(list, out);

    // Dialog
    WCreateMarkerDialog dlg(title, out, pAdjacent, list);
    ui_root->centerWidget(dlg);
    int result = ui_root->runModal(dlg);
    if (result != Cancel) {
        if (pAdjacent) {
            *pAdjacent = (result == Adjacent);
        }
        out = dlg.getResult();
        return true;
    } else {
        return false;
    }
}
