/**
  *  \file client/dlg-supplies.cc
  *  \brief Supply Sale Dialog
  *
  *  This module contains the user interface for the "sell supplies"
  *  function.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/dlg-supplies.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/undo.h"
#include "game/actions/planetsupplies.h"
#include "ui/widgets/button.h"
#include "client/widgets/helpwidget.h"
#include "util/misc.h"
#include "ui/widgets/quit.h"

const int WIDTH = 400;

/************************** WSellSuppliesDialog **************************/

/** \class WSellSuppliesDialog
    \brief Supply Sale Dialog

    This is the standard "sell supplies" dialog.

    When run, and the user chose "OK", the dialog will call
    GPlanetSellSuppliesAction::setAmount() and return nonzero; the
    caller must commit the action. When the user chooses "Cancel",
    this will return zero; the caller must cancel the transaction. */

/** Constructor. Constructs a complete dialog. */
WSellSuppliesDialog::WSellSuppliesDialog(GPlanetSellSuppliesAction& act)
    : UIWindow(_("Sell Supplies"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      select(100, 0, act.getMaximumAmount(), 10), action(act)
{
    init();
}

void
WSellSuppliesDialog::init()
{
    setLayout(UIVBoxLayout::instance5);
    add(holder.add(new UIMultilineStatic(FONT_NORMAL,
                                    format(_("You have %d kt supplies. You'll get 1 mc per kiloton. "
                                             "Remember that PCC automatically sells supplies when needed.\n"
                                             "Enter amount to sell:"),
                                           numToString(action.getMaximumAmount())),
                                    WIDTH)));
    add(select);

    UIGroup& g = holder.add(new UIGroup(UIHBoxLayout::instance5));
    add(g);

    UIButton& ok = holder.add(new UIButton(1000, _("OK"), SDLK_RETURN));
    ok.sig_fired.add(this, &WSellSuppliesDialog::onEnter);
    g.add(ok);

    UIButton& all = holder.add(new UIButton(1001, UIKeyString(_("All but..."))));
    all.sig_fired.add(this, &WSellSuppliesDialog::onAll);
    g.add(all);

    UIButton& esc = holder.add(new UIButton(1002, _("Cancel"), SDLK_ESCAPE));
    esc.sig_fired.add(this, &WSellSuppliesDialog::onCancel);
    g.add(esc);

    g.add(holder.add(new UISpacer()));

    g.add(holder.add(new UIButton(_("Help"), 'h')));
    add(holder.add(new WHelpWidget("pcc2:sellsup")));
    add(holder.add(new UIQuit(false /* xref onCancel */)));

    pack();
}

WSellSuppliesDialog::~WSellSuppliesDialog()
{ }

/** Handler for "OK" button. */
void
WSellSuppliesDialog::onEnter(int)
{
    action.setAmount(select.getValue());
    stop(true);
}

/** Handler for "Sell all but" button. */
void
WSellSuppliesDialog::onAll(int)
{
    if (select.getValue() == 0 && action.getMaximumAmount() != 0)
        if (!UIMessageBox(_("Do you really want to sell all supplies?"), _("Sell Supplies")).doYesNoDialog())
            return;
    action.setAmount(action.getMaximumAmount() - select.getValue());
    stop(true);
}

/** Handler for "Cancel" button. */
void
WSellSuppliesDialog::onCancel(int)
{
    stop(false);
}

/** Main entry. Constructs the dialog, runs it, and applies it to
    planet p. */
void
WSellSuppliesDialog::doDialog(GPlanet& p)
{
    if (p.getCargo(el_Supplies) == 0) {
        messageBox(_("You do not have any supplies on this planet."),
                   _("Sell Supplies"));
        return;
    }

    GPlanetSellSuppliesAction act(p);
    WSellSuppliesDialog dlg(act);

    ui_root->centerWidget(dlg);
    ui_root->add(dlg);
    int i = dlg.run();
    ui_root->close(dlg);

    if (i)
        act.doCommit();
    else
        act.doCancel();
}

/************************ WUndoSellSuppliesDialog ************************/

/** \class WUndoSellSuppliesDialog
    \brief Undo Supply Sale (buy) dialog

    This dialog prompts the user for a number of supplies to buy, and
    performs the order. Unlike many other dialogs, this one bypasses
    the cargo reservation system: it takes the user's input, and
    validates it afterwards. If the user enters an invalid value, it
    also updates the prompt accordingly. */

/** Constructor. Builds the whole dialog.
    \param turn Turn to work on
    \param pid  Planet Id */
WUndoSellSuppliesDialog::WUndoSellSuppliesDialog(GGameTurn& turn, int pid)
    : UIWindow(_("Undo Supply Sale"), BLUE_WINDOW, COLOR_GRAY),
      select(100, 0, 0, 10),
      introtext(FONT_NORMAL, string_t(), WIDTH),
      turn(turn),
      pid(pid),
      maximum(0)
{
    init();
}

void
WUndoSellSuppliesDialog::init()
{
    setLayout(UIVBoxLayout::instance5);

    /* This will also fill in the 'maximum' field and place nice
       text in the controls */
    computeRange();

    /* Now build the dialog */
    add(introtext);
    add(select);

    UIGroup& g = holder.add(new UIGroup(UIHBoxLayout::instance5));
    add(g);

    UIButton& ok = holder.add(new UIButton(1000, _("OK"), SDLK_RETURN));
    ok.sig_fired.add(this, &WUndoSellSuppliesDialog::onEnter);
    g.add(ok);

    UIButton& esc = holder.add(new UIButton(1002, _("Cancel"), SDLK_ESCAPE));
    esc.sig_fired.add(this, &WUndoSellSuppliesDialog::onCancel);
    g.add(esc);

    g.add(holder.add(new UISpacer()));

    g.add(holder.add(new UIButton(_("Help"), 'h')));
    add(holder.add(new WHelpWidget("pcc2:sellsup")));
    add(holder.add(new UIQuit(false /* xref onCancel */)));

    pack();
}

WUndoSellSuppliesDialog::~WUndoSellSuppliesDialog()
{ }

void
WUndoSellSuppliesDialog::onEnter(int)
{
    int32 entered_value = select.getValue();
    computeRange();

    if (entered_value != select.getValue()) {
        /* This means the situation changed in the meantime. It might
           even mean we have to re-layout */
        pack();
    } else {
        /* No change, so be it */
        GPlanet& p = turn.getCurrentUniverse().getPlanet(pid);
        p.changeCargoRaw(el_Supplies, +entered_value);
        p.changeCargoRaw(el_Money,    -entered_value);
        turn.getCurrentUniverse().doScreenUpdates();
        stop(1);
    }
}

void
WUndoSellSuppliesDialog::onCancel(int)
{
    stop(0);
}

void
WUndoSellSuppliesDialog::computeRange()
{
    /* Compute undo information */
    GUndoInformation undo;
    undo.computeInformation(turn, pid);

    /* Number of supplies that can be bought is limited by the money
       we currently have, of course */
    int32 supplies = undo.getSuppliesAllowedToBuy();
    int32 money    = turn.getCurrentUniverse().getPlanet(pid).getCargo(el_Money);

    int32 new_maximum = supplies > money ? money : supplies;

    if (new_maximum != maximum) {
        maximum = new_maximum;

        /* Update the MultilineStatic. */
        introtext.setText(format(_("You can buy up to %d supplies. (This is the amount of supplies "
                                   "you already sold this turn but have not used otherwise.)\n"
                                   "Enter amount to buy:"),
                                 numToString(maximum)));

        /* Update the decimal selector */
        select.setRange(0, maximum);
    }
}

/** Main entry point. This creates the dialog and runs it. */
void
WUndoSellSuppliesDialog::doDialog(GGameTurn& turn, int pid)
{
    /* We build the complete dialog first and then check whether we
       actually need it. This will create several widgets and thus is
       a little more expensive than it needs to be, but it is much
       simpler. */
    WUndoSellSuppliesDialog dlg(turn, pid);
    if (dlg.maximum == 0) {
        messageBox(_("You cannot buy supplies. Either you have not yet sold any this turn, "
                     "or you have already spent the money."),
                   _("Undo Supply Sale"));
    } else {
        ui_root->centerWidget(dlg);
        ui_root->add(dlg);
        ui_root->run();
        ui_root->close(dlg);
    }
}
