/**
  *  \file client/playvcropts.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/playvcropts.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "util/misc.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "arch/platform.h"
#include "io/dirs.h"
#include "util/recode.h"

/** Initialize. */
PlayvcrOptionParser::PlayvcrOptionParser()
    : mode(mUnset), num_nonoptions(0), battle_nr(0)
{
    static const Option options[] = {
        { "help", help, 0, false,
          N_(" -help              Show help screen") },
        { "dir", setDirectoryMode, 0, false,
          N_(" -dir               Start with directory selection") },
        { "charset", setCharset, 0, true,
          N_(" -charset EXT[:INT] Set character set translation") },
        { "sim", setSimulatorMode, 0, false,
          N_(" -sim               Run simulator, not VCR") },
    };
    addOptionSet(options, countof(options));
}

/** Handle Non-Options. The first non-option is the file name, the
    second (optional) is the directory. A battle number may also
    be specified. */
void
PlayvcrOptionParser::handleNonoption(const string_t& str)
{
    int tmp;
    if (battle_nr == 0 && stringValue(str, tmp) && tmp > 0) {
        battle_nr = tmp;
    } else {
        if (num_nonoptions >= MAX_NON_OPTIONS) {
            handleError(_("Too many parameters specified."));
        } else {
            nonopts[num_nonoptions++] = str;
        }
    }
}

/** Handle "-help" option: show help screen. */
void
PlayvcrOptionParser::help(StandardCommandLineParser& par,
                          const StandardCommandLineParser::Option&)
{
    console << _("Usage: playvcr [-options] vcrX.dat [gamedir] [battlenr]\n"
                 "       playvcr [-options] -sim gamedir\n"
                 "       playvcr [-options] file.ccb [gamedir]\n"
                 "\n"
                 "Valid options:\n");
    par.printHelp(console);
    console << _("\n(c) 2001-2022 Stefan Reuther <Streu@gmx.de>\n");
    Platform::showOutput();
    std::exit(0);
}

/** Handle "-charset" option: set character set translation. */
void
PlayvcrOptionParser::setCharset(StandardCommandLineParser& par,
                                const StandardCommandLineParser::Option&)
{
    string_t csName = par.getArg();
    CharacterSet cs = CharacterSet::getCharsetByName(csName);
    if (!cs.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), csName));
    setGameCharacterSet(cs);
}

/** Handle "-dir" option. */
void
PlayvcrOptionParser::setDirectoryMode(StandardCommandLineParser& self, const Option&)
{
    static_cast<PlayvcrOptionParser&>(self).setMode(mChooseFile);
}

/** Handle "-sim" option. */
void
PlayvcrOptionParser::setSimulatorMode(StandardCommandLineParser& self, const Option&)
{
    static_cast<PlayvcrOptionParser&>(self).setMode(mSimulate);
}

/** Error handler. Instead of throwing an exception, show a message
    and bail out. */
void
PlayvcrOptionParser::handleError(const string_t& str)
{
    console.write(LOG_ERROR, str);
    console.write(LOG_ERROR, _("Try `playvcr --help' for help."));
    Platform::showOutput();
    std::exit(1);
}

/** Parsing finished. Postprocess input. */
void
PlayvcrOptionParser::parseFinished()
{
    if (mode == mUnset) {
        if (getNumNonoptions() == 0) {
            /* No parameters given: choose */
            setMode(mChooseFile);
            setDirName(Directory::getWorkingDirectory());
        } else {
            /* One or two parameters: play file. */
            setMode(mPlayFile);
        }
    }

    /* Postprocess */
    if (mode == mPlayFile) {
        /* Play file: fill in "directory" field */
        if (getNumNonoptions() == 1) {
            setDirName(Directory::getDirName(getFileName()));
        }
    } else if (mode == mChooseFile || mode == mSimulate) {
        /* Simulate or choose */
        if (getNumNonoptions() == 0) {
            setDirName(Directory::getWorkingDirectory());
        } else if (getNumNonoptions() == 1) {
            setDirName(getFileName());
        } else {
            handleError(_("Only one parameter expected when `-dir' or `-sim' is specified."));
        }
    } else {
        /* can't happen */
    }
}
