/**
  *  \file client/screens/planetscreen.cc
  *  \brief Planet Screen
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/planetscreen.h"
#include "client/cscreen.h"
#include "client/dialogs/objselect.h"
#include "client/tiles/planetoverview.h"
#include "client/wscanchart.h"
#include "game/global.h"
#include "game/objsel.h"
#include "int/if/planetif.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "ui/commands.h"

namespace {
    class WPlanetScreen : public WControlScreen {
     public:
        WPlanetScreen();
        ~WPlanetScreen();
        void update();
        virtual void enumContexts(ptr_vector<IntContext>& ctx);
    };
}

/** Choose a planet. Displays the dialog if necessary and lets the
    user choose a planet.
    \return planet Id, zero if no planet or cancelled */
int
choosePlanet()
{
    if (current_planet_type.isEmpty()) {
        messageBox(_("You do not have any planets. "
                     "Unload colonists from a starship to a planet to colonize it."),
                   _("Planets"));
        return 0;
    } else if (current_planet_type.isUnit()) {
        return current_planet_selection.getCurrentIndex();
    } else {
        GSimpleObjectSelection select;
        select.initFrom(current_planet_selection);
        WPlanetOverviewTile bot(select);
        WObjectSelectionDialog dlg(_("Select Planet"), "pcc2:unitsel", select);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter) {
            return select.getCurrentIndex();
        } else {
            return 0;
        }
    }
}

/***************************** WPlanetScreen *****************************/

/** \class WPlanetScreen
    \brief Planet Control Screen

    Standard control screen for a player's planet. */

/** Constructor. */
WPlanetScreen::WPlanetScreen()
    : WControlScreen(current_planet_selection, WScreenId::PlanetScreen)
{
    createTileLayout(*this, "PLANETSCREEN", current_planet_selection);
    addChart(holder.add(new WScannerChartWidget()));
    setKeymap(IntKeymap::getKeymapByName("PLANETSCREEN"));
}

WPlanetScreen::~WPlanetScreen()
{ }

void
WPlanetScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(getObjectSelection().getCurrentObject())) {
        getScanner().getChart()->setCenterPosition(mo->getPos());
        getScanner().getChart()->setScannerOrigin(mo->getPos());
        getScanner().getChart()->setScannerTarget(mo->getPos());
    }
}

void
WPlanetScreen::enumContexts(ptr_vector<IntContext>& ctx)
{
    ctx.push_back(new IntPlanetContext(getObjectSelection().getCurrentIndex()));
}

void
doPlanetScreen()
{
    WPlanetScreen().runScreen();
}
