/**
  *  \file game/actions/baseammo.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/baseammo.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/undo.h"
#include "game/global.h"

/** Constructor.
    \param pay       who will pay for the ammo, must be a planet
    \param receiver  who will receive the ammo, can be any container
    \pre we are playing pay and receiver, and planet pay has a base */
GStarbaseBuildAmmoAction::GStarbaseBuildAmmoAction(GPlanetTransfer& pay,
                                                   GCargoContainer& receiver)
    : GCargoCostTransaction(pay), pay(pay), receiver(receiver)
{
    for (std::size_t i = 0; i < countof(added); ++i) {
        added[i] = 0;
        allowed_to_sell[i] = 0;
    }
}

GStarbaseBuildAmmoAction::~GStarbaseBuildAmmoAction()
{
    if (!isFinished())
        doCancel();
}

void
GStarbaseBuildAmmoAction::setUndoInformation(const GUndoInformation& info)
{
    for (size_t i = 0; i < countof(allowed_to_sell); ++i) {
        allowed_to_sell[i] = info.getAmmoAllowedToSell(getCargoTypeFromIndex(i+1));
    }
}

/** Set undo information from GGameTurn.
    If the GGameTurn matches our planet, uses it to derive undo information. */
void
GStarbaseBuildAmmoAction::setDefaultUndoInformation(const GGameTurn& trn)
{
    GPlanet& planet = getPlanet();
    if (&trn.getCurrentUniverse().getPlanet(planet.getId()) == &planet) {
        GUndoInformation undo;
        undo.computeInformation(trn, planet.getId());
        setUndoInformation(undo);
    }
}

/** Convert index into cargo type. This class uses indexes
    [min_index,max_index] to identify the ammo types. This function
    returns the corresponding cargo type. */
GCargoType
GStarbaseBuildAmmoAction::getCargoTypeFromIndex(int index)
{
    if (index > 0 && index <= NUM_TORPS) {
        return (GCargoType(el_Torps + index-1));
    } else if (index == max_index) {
        return el_Fighters;
    } else {
        ASSERT(0);
        return el_Neutronium; // silence compiler
    }
}

/** Check whether receiver can have the specified ammo type.
    \param index type index, [min_index,max_index]
    \return true iff receiver can carry that type */
bool
GStarbaseBuildAmmoAction::canHave(int index) const
{
    return receiver.canHaveCargo(getCargoTypeFromIndex(index));
}

/** Get current amount of specified ammo type.
    \param index type index, [min_index,max_index]
    \return amount of specified type on target */
int32_t
GStarbaseBuildAmmoAction::getAmount(int index) const
{
    return receiver.getCargo(getCargoTypeFromIndex(index));
}

/** Set current amount on receiver.
    \param index type index, [min_index,max_index]
    \param amount new amount

    The new amount can be any value. If it is too small, or if the
    receiver cannot carry that much, it is adjusted. If it exceeds the
    payers resources, we will go into an error state until the amount
    is reduced again. */
void
GStarbaseBuildAmmoAction::setAmount(int index, int32 amount)
{
    // difference to total amount
    int32_t diff = amount - getAmount(index);

    // we can't take away more than we have
    if (diff < 0 && -diff > added[index-1] + allowed_to_sell[index-1])
        diff = -(added[index-1] + allowed_to_sell[index-1]);

    // virtually add it to underlying object
    added[index-1] += receiver.changeCargo(getCargoTypeFromIndex(index), diff, true);

    // re-compute cost
    GCost sum;
    for (int i = 1; i <= NUM_TORPS; ++i)
        if (added[i-1] != 0)
            sum += getTorp(i).getTorpedo().getCost() * added[i-1];

    if (added[NUM_TORPS] != 0)
        sum += getFighterForPlayer(pay.getPlanet().getOwner()).getCost() * added[NUM_TORPS];

    setCost(sum);
}

/** Limit current amount. A negative amount is limited such that we do not
    sell items we didn't buy. A positive amount is limited by the available
    resources.
    \param index type index, [min_index,max_index]
    \param amount new amount
    \return adjusted amount */
int32_t
GStarbaseBuildAmmoAction::limitAmount(int index, int32_t amount) const
{
    if (amount < 0) {
        // We cannot remove more than we added
        if (-amount > added[index-1] + allowed_to_sell[index-1])
            amount = -(added[index-1] + allowed_to_sell[index-1]);
    } else {
        // Compute item cost
        GCost item_cost = (index == max_index
                           ? getFighterForPlayer(pay.getPlanet().getOwner()).getCost()
                           : getTorp(index).getTorpedo().getCost());

        // Convert remainder into cost
        GCost remainder;
        remainder.set(el_Tritanium,  getRemainingCargo(el_Tritanium));
        remainder.set(el_Duranium,   getRemainingCargo(el_Duranium));
        remainder.set(el_Molybdenum, getRemainingCargo(el_Molybdenum));
        remainder.set(el_Supplies,   getRemainingCargo(el_Supplies));
        remainder.set(el_Money,      getRemainingCargo(el_Money));

        amount = remainder.limitAmount(amount, item_cost);
    }
    return amount;
}

void
GStarbaseBuildAmmoAction::commit()
{
    GCargoCostTransaction::commit();
    receiver.mustCommit();
}

void
GStarbaseBuildAmmoAction::cancel()
{
    GCargoCostTransaction::cancel();
    receiver.mustCancel();
}

/** Get required tech upgrade. The tech upgrade must be done before using this
    transaction, in a separate transaction.
    \param index type index, [min_index,max_index]
    \return torp tech level that must be acquired to build the
    requested item. 0 if no tech upgrade needed. */
int
GStarbaseBuildAmmoAction::getTechUpgradeNeeded(int index) const
{
    if (index > 0 && index <= NUM_TORPS) {
        int tech = getTorp(index).getTech();
        if (tech > pay.getPlanet().getBaseTechLevel(tl_Torpedo))
            return tech;
    }
    return 0;
}
