/**
  *  \file game/classicvcr.h
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CLASSICVCR_H
#define PCC_V2_GAME_CLASSICVCR_H

#include "game/vcrdb.h"
#include "game/vcr.h"
#include "cpluslib/ptrvector.h"

class Stream;

/****************************** Classic VCR ******************************/

enum GClassicVcrType {
    vcr_Host,               ///< Host.
    vcr_Unknown,            ///< Unknown.
    vcr_UnknownPHost,       ///< Unknown brand of PHost.
    vcr_PHost2,             ///< PHost 2.
    vcr_PHost3,             ///< PHost 3 or PHost 4 without extensions.
    vcr_PHost4,             ///< PHost 4 with extensions.
    vcr_NuHost              ///< NuHost.
};

inline bool
isPHost(GClassicVcrType t)
{
    switch (t) {
     case vcr_Unknown:      break;
     case vcr_Host:         break;
     case vcr_UnknownPHost: return true;
     case vcr_PHost2:       return true;
     case vcr_PHost3:       return true;
     case vcr_PHost4:       return true;
     case vcr_NuHost:       break;
    }
    return false;
}

/** Entry in a GClassicVcrDatabase. */
class GClassicVcrEntry : public GVcrEntry {
    int             status;
    GClassicVcrType type;
    uint16_t        capabilities;
    GClassicVcr     vcr;
    GClassicVcr     result;

 public:
    explicit GClassicVcrEntry(const GClassicVcr& vcr);
    ~GClassicVcrEntry();

    // Initialisation:
    void setType(GClassicVcrType type, uint16_t capabilities);
    GClassicVcrType getType() const;
    uint16_t getCapabilities() const;

    // GVcrEntry:
    int getNumObjects() const;
    GVcrObject& getObject(int slot, bool after);
    int getOutcome(int slot);
    Playability getPlayability();
    void prepareResult(int resultLevel, ProgressMonitor& pm);
    string_t getAlgorithmName(bool translate) const;
    bool isESBActive();

    // GClassicVcrEntry:
    void setResultFromPlayer(VcrPlayer& player);
    Ptr<VcrPlayer> createPlayer();
    static Ptr<VcrPlayer> createPlayerForType(GClassicVcrType type);
    void applyClassicLimits();

    /** Get combat recording in "raw" format. */
    const GClassicVcr& getVcr() const
        { return vcr; }
    /** Get result of combat in "raw" format. */
    const GClassicVcr& getResult() const
        { return result; }
    /** Get status of fight, if known. */
    int getStatus() const
        { return status; }

    /** Scores. */
    enum {
        EarnedBuildMillipointsMin,
        EarnedBuildMillipointsMax,
        EarnedExperience,
        EarnedTonsDestroyed,
        NumScores
    };
    void computeScores(int32_t scores[NumScores], int side) const;
};

class GClassicVcrDatabase : public GVcrDatabase {
 public:
    GClassicVcrDatabase();
    GClassicVcrDatabase(Stream& file);
    ~GClassicVcrDatabase();

    void load(Stream& file);

    // GVcrDatabase:
    virtual int getNumBattles() const;
    virtual GClassicVcrEntry& getBattle(int nr);

    GClassicVcrEntry& addEntry(const GClassicVcr& vcr);

 private:
    ptr_vector<GClassicVcrEntry> content;
};

#endif
