/**
  *  \file gfx/gfxutil.h
  *  \brief Graphics Utilities
  *
  *  Utility classes used for implementing the graphics and UI system.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFXUTIL_H
#define PCC_V2_GFX_GFXUTIL_H

#include <iosfwd>
#include <list>
#include <SDL_video.h>
#include "cpluslib/types.h"
#include "gfx/gfxtypes.h"

class GfxCanvas;

/** Fill Pattern. An 8x8 pattern usable to fill a rectangle.
    Consists of 8 bytes (one per line, first byte is topmost),
    the most-significant bit of each byte is leftmost. */
class GfxFillPattern {
    uint8 pat[8];  ///< Pattern data.
 public:
    GfxFillPattern() throw();
    GfxFillPattern(const uint8* init) throw();
    GfxFillPattern(uint8 val) throw();

    bool isBlank() const throw();
    bool isBlack() const throw();

    GfxFillPattern& shiftLeft(int amount) throw();
    GfxFillPattern& shiftRight(int amount) throw();
    GfxFillPattern& shiftUp(int amount) throw();
    GfxFillPattern& shiftDown(int amount) throw();

    GfxFillPattern& operator |= (const GfxFillPattern& rhs) throw();
    GfxFillPattern& operator |= (uint8 value) throw();
    GfxFillPattern& operator &= (const GfxFillPattern& rhs) throw();
    GfxFillPattern& operator &= (uint8 value) throw();
    GfxFillPattern& operator ^= (const GfxFillPattern& rhs) throw();
    GfxFillPattern& operator ^= (uint8 value) throw();

    GfxFillPattern& invert() throw();

    GfxFillPattern& flipHorizontal() throw();
    GfxFillPattern& flipVertical() throw();

    /** Read/write access to line i. */
    uint8& operator[](unsigned i)
        { return pat[i & 7]; }
    /** Read access to line i. */
    uint8 operator[](unsigned i) const
        { return pat[i & 7]; }
};

/** Point. This object contains a point in the X/Y plane. This is a
    plain old structure, don't hesitate to modify it directly. */
struct GfxPoint {
    int x;    ///< X coordinate.
    int y;    ///< Y coordinate.

    /** Create uninitialized(!) point. */
    GfxPoint()
        { }
    /** Create point at specified X/Y. */
    GfxPoint(int ax, int ay)
        : x(ax), y(ay) { }
    /** Compare two points for equality. */
    bool operator==(GfxPoint rhs) const
        { return x == rhs.x && y == rhs.y; }
    /** Compare two points for inequality. */
    bool operator!=(GfxPoint rhs) const
        { return x != rhs.x || y != rhs.y; }
    /** Move point by relative distance.
        \param dx,dy Distance
        \return moved point (does not change *this)! */
    GfxPoint movedBy(int dx, int dy) const
        { return GfxPoint(x + dx, y + dy); }
    /** Move point by relative distance.
        \param pt Specifies distance
        \return moved point (does not change *this)! */
    GfxPoint movedBy(GfxPoint pt) const
        { return GfxPoint(x + pt.x, y + pt.y); }
    /** Scale point coordinates.
        \param sx Multiply x by this value
        \param sy Multiply y by this value
        \return scaled point (does not change *this)! */
    GfxPoint scaledBy(int sx, int sy) const
        { return GfxPoint(x * sx, y * sy); }
};

/** Rectangle. An SDL_Rect or a faster implementation. Represents
    rectangles as position plus extent. This offers some useful
    primitives to work with rectangles. Otherwise, this is a plain old
    structure and you are invited to modify it directly. */
class GfxRect : public SDL_Rect {
 public:
    GfxRect() throw();
    GfxRect(const SDL_Rect& r) throw();
    GfxRect(int x, int y, int w, int h) throw();
    explicit GfxRect(GfxCanvas& can);
    explicit GfxRect(GfxPoint pt) throw();
    GfxRect(int x, int y, GfxPoint pt) throw();
    GfxRect(GfxPoint pt, int w, int h) throw();
    GfxRect(GfxPoint pt, GfxPoint ex) throw();

    void clipSize(int cw, int ch) throw();
    void clipRect(const SDL_Rect& r) throw();
    void include(const SDL_Rect& r) throw();
    void include(GfxPoint pt) throw();
    bool contains(const GfxRect& r) const throw();
    bool operator==(const SDL_Rect& rhs) const throw();
    bool operator!=(const SDL_Rect& rhs) const throw();
    GfxPoint moveTo(GfxPoint where) throw();

    /** Check whether this rectangle has a non-zero extent. */
    bool exists() const throw()
        { return w > 0 && h > 0; }
    /** Compute intersection. Same as clipRect(), see there. */
    void intersect(const SDL_Rect& r) throw()
        { clipRect(r); }
    /** Check whether rectangle contains a given point. */
    bool contains(int px, int py) const
        {
            px -= x;
            py -= y;
            return px >= 0 && py >= 0 && unsigned(px) < w && unsigned(py) < h;
        }
    /** Check whether rectangle contains a given point. */
    bool contains(GfxPoint pt) const throw()
        { return contains(pt.x, pt.y); }
    /** Move to point. \overload */
    GfxPoint moveTo(int px, int py) throw()
        { return moveTo(GfxPoint(px, py)); }
    /** Move by relative distance. The distance is given in the
        components of a point. */
    void moveBy(GfxPoint dist)
        { x += dist.x; y += dist.y; }
    /** Move by relative distance. The distance is given by two
        integers. */
    void moveBy(int dx, int dy)
        { x += dx; y += dy; }
    /** Move by relative distance, reverse direction. */
    void moveByRev(GfxPoint dist)
        { x -= dist.x; y -= dist.y; }
    /** Move by relative distance, reverse direction. */
    void moveByRev(int dx, int dy)
        { x -= dx; y -= dy; }
    /** Grow rectangle. This enlarges the rectangle by \c dx columns
        to the left and right, and by \c dy lines at top and bottom
        (for a total growth of twice the given amount, of course). Use
        negative values to shrink. */
    void grow(int dx, int dy)
        { x -= dx; y -= dy; w += 2*dx; h += 2*dy; }
    /** Grow rectangle in one direction. This enlarges the rectangle
        by \c dx columns to the right and \c dy lines to the bottom,
        without affecting its origin. */
    void growSize(int dx, int dy)
        { w += dx; h += dy; }
    /** Grow rectangle in one direction. \overload */
    void growSize(GfxPoint pt)
        { w += pt.x; h += pt.y; }
    /** Get center point of rectangle. */
    GfxPoint getCenter() const
        { return GfxPoint(x + w/2, y + h/2); }
    /** Check whether this rectangle intersects another one. */
    bool isIntersecting(GfxRect r) const
        {
            r.clipRect(*this);
            return r.exists();
        }
    /** Get bottom coordinate. */
    int getBottomY() const
        { return y + h; }
    /** Get right coordinate. */
    int getRightX() const
        { return x + w; }
    /** Get top-left point. */
    GfxPoint getTopLeft() const
        { return GfxPoint(x, y); }
    /** Get bottom-right point. */
    GfxPoint getBottomRight() const
        { return GfxPoint(getRightX(), getBottomY()); }
    /** Get size of rectangle. */
    GfxPoint getSize() const
        { return GfxPoint(w, h); }

    void moveInto(const GfxRect& other) throw();
    void centerWithin(const GfxRect& other) throw();
    void moveToEdge(const GfxRect& other, int x_pos, int y_pos, int offset) throw();

    void consumeX(int pix);
    GfxRect splitX(int pix);

    void consumeY(int pix);
    GfxRect splitY(int pix);
};

class GfxRectSet {
    std::list<GfxRect> the_list;

    static void doSubtract(const GfxRect& r, std::list<GfxRect>& l2, std::list<GfxRect>& other);
 public:
    typedef std::list<GfxRect>::const_iterator iterator;

    GfxRectSet();
    GfxRectSet(const GfxRect& r);

    void intersect(const GfxRect& r);
    void add(const GfxRect& r);
    void remove(const GfxRect& r);
    void clear()
        { the_list.clear(); }

    bool contains(int x, int y) const;
    bool contains(const GfxPoint& r) const
        { return contains(r.x, r.y); }
    bool contains(const GfxRect& r) const;

    GfxRect getBoundingRectangle() const;

    /** Iterator interface: get iterator to first element. */
    iterator begin() const
        { return the_list.begin(); }
    /** Iterator interface: get iterator to last element. */
    iterator end() const
        { return the_list.end(); }
    /** Check whether this set is empty. */
    bool empty() const
        { return the_list.empty(); }
    /** Get number of rectangles in this rectangle set. */
    std::list<GfxRect>::size_type getNumRectangles() const
        { return the_list.size(); }
};

extern const GfxFillPattern SOLID_FILL, GRAY50_FILL, GRAY25_FILL,
    GRAY50_FILL2, LTSLASH_FILL;

std::ostream& operator<<(std::ostream&, const GfxRect&);
std::ostream& operator<<(std::ostream&, GfxPoint);

/******************************** Inlines ********************************/

/** Construct empty GfxRect.
    \post contains(p) == false for every p */
inline
GfxRect::GfxRect() throw()
{
    x = y = w = h = 0;
}

/** Construct GfxRect from SDL_Rect. */
inline
GfxRect::GfxRect(const SDL_Rect& r) throw()
    : SDL_Rect(r)
{ }

/** Construct GfxRect from explicit parameters.
    \param x left
    \param y top
    \param w width
    \param h height */
inline
GfxRect::GfxRect(int x, int y, int w, int h) throw()
{
    this->x = x;
    this->y = y;
    this->w = w;
    this->h = h;
}

/** Construct rectangle with given size. The rectangle is located at (0,0).
    \param pt Specifies size of rectangle */
inline
GfxRect::GfxRect(GfxPoint pt) throw()
{
    this->x = 0;
    this->y = 0;
    this->w = pt.x;
    this->h = pt.y;
}

/** Construct rectangle with given size at position.
    \param x,y Origin
    \param pt  Specifies size of rectangle */
inline
GfxRect::GfxRect(int x, int y, GfxPoint pt) throw()
{
    this->x = x;
    this->y = y;
    this->w = pt.x;
    this->h = pt.y;
}

/** Construct rectangle at given position.
    \param pt Origin
    \param w,h Size */
inline
GfxRect::GfxRect(GfxPoint pt, int w, int h) throw()
{
    this->x = pt.x;
    this->y = pt.y;
    this->w = w;
    this->h = h;
}

/** Construct rectangle with position and size.
    \param pt Origin
    \param ex Specifies size of rectangle */
inline
GfxRect::GfxRect(GfxPoint pt, GfxPoint ex) throw()
{
    this->x = pt.x;
    this->y = pt.y;
    this->w = ex.x;
    this->h = ex.y;
}

#endif
