/**
  *  \file int/expr.h
  *  \brief Interpreter: Expression Parse Trees
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_EXPR_H
#define PCC_V2_INT_EXPR_H

#include "cpluslib/object.h"
#include "int/bytecode.h"
#include "cpluslib/ptrvector.h"

class IntValue;

/** Compilation context. Contains options in effect for current compilation. */
struct IntCompilationContext {
    enum {
        /** Expressions: If set, string operations are case-blind ("NC" opcodes). */
        CaseBlind = 1,
        /** Expressions: If set, code will execute directly in the BCO's context; no 'With' or 'ForEach' active.
            This allows the code generator to generate 'pushloc' instead of 'pushvar' instructions
            for local names. */
        LocalContext = 2,
        /** Expressions: If this is set along with LocalContext, the parent of the BCO's context will be the
            shared variables. */
        AlsoGlobalContext = 4,
        /** Statements: if set, expressions are statements, and that's it. The expression will be compiled
            into its side-effect, discarding the result. If clear, we want expression results. */
        ExpressionsAreStatements = 8,
        /** Statements: restrict to one-liners. If set, multi-line blocks are refused. */
        RefuseBlocks = 16,
        /** Statements: accept terminators. If set, terminators are reported to the caller of the compiler;
            EOF is an error. If clear, terminators are not expected and cause an error, EOF terminates
            compilation. */
        WantTerminators = 32,
        /** Statements: linear execution until here. If set, it is guaranteed that the statements within this
            block are guaranteed to be executed in their entirety, linearly once from top to bottom. */
        LinearExecution = 64
    };
    uint32_t compilation_flags;

    IntCompilationContext()
        : compilation_flags(CaseBlind)
        { }

    IntCompilationContext& withFlags(uint32_t flags)
        { compilation_flags |= flags; return *this; }

    IntCompilationContext& withoutFlags(uint32_t flags)
        { compilation_flags &= ~flags; return *this; }

    bool hasFlags(uint32_t which) const
        { return (compilation_flags & which) != 0; }
};

/** Basic expression node. */
class IntExprNode : public Object {
    IntExprNode& operator=(const IntExprNode&);
    IntExprNode(const IntExprNode&);

 public:
    /** Constructor. */
    IntExprNode()
        { }

    /** Compile effect of this expression. Execution stack will not be changed.
        Default is to do compileValue and drop result. */
    virtual void compileEffect(IntBytecodeObject& bco, const IntCompilationContext& cc);

    /** Compile value of this expression. Result will be on stack. */
    virtual void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc) = 0;

    /** Store into this expression. Store stack-top into this expression (and leave it there). */
    virtual void compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs) = 0;

    /** Compile as condition. Generates a jump to \c ift if the expression is true,
        to \c iff if the expression is false or empty. Does not change the stack.
        Default is to compile value and generate two jumps. */
    virtual void compileCondition(IntBytecodeObject& bco, const IntCompilationContext& cc, IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff);

    /** Compile read-modify-write cycle, "read" half. Result will be top-of-stack; state information can be below. */
    virtual void compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc) = 0;

    /** Compile read-modify-write cycle, "write" half. Updated value
        is on stack; must remain top-of-stack, although state
        information must be removed. */
    virtual void compileWrite(IntBytecodeObject& bco, const IntCompilationContext& cc) = 0;
};

/** Generic r-value expression. This one refuses to be assigned-to. */
class IntRValueExprNode : public IntExprNode {
 public:
    void compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs);
    void compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileWrite(IntBytecodeObject& bco, const IntCompilationContext& cc);
};

/** Literal expression node. Generates code to return a literal value. */
class IntLiteralExprNode : public IntRValueExprNode {
    IntValue* value;

 public:
    IntLiteralExprNode();
    ~IntLiteralExprNode();
    void setValue(IntValue* value) throw();
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
};

/** Basic rvalue node. This is an internal node with up to three child nodes.
    Code generation is implemented in derived classes. */
class IntSimpleRValueExprNode : public IntRValueExprNode {
 protected:
    IntExprNode* a;
    IntExprNode* b;
    IntExprNode* c;

 public:
    IntSimpleRValueExprNode();
    ~IntSimpleRValueExprNode();
    void setUnary(IntExprNode* a) throw();
    void setBinary(IntExprNode* a, IntExprNode* b) throw();
    void setTernary(IntExprNode* a, IntExprNode* b, IntExprNode* c) throw();
};

/** Simple expression node. Combines up to three parameters using a single opcode. */
class IntSimpleExprNode : public IntSimpleRValueExprNode {
    IntOpcode::Major major : 8;
    uint8_t minor;

 public:
    IntSimpleExprNode(IntOpcode::Major major, uint8_t minor);
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    bool is(IntOpcode::Major major, uint8_t minor) const
        { return this->major == major && this->minor == minor; }
};

/** Case-sensitive expression node. Generates a binary operation with the specified
    minor opcode when in case-sensitive mode, with minor+1 when case-insensitive.
    \todo Can we merge this somehow with IntCaseFunctionNode? */
class IntCaseExprNode : public IntSimpleRValueExprNode {
    uint8_t minor;

 public:
    IntCaseExprNode(uint8_t minor);
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    IntExprNode* convertToAssignment();
};

/** Assignment node. Assigns node b to node a. */
class IntAssignmentNode : public IntSimpleRValueExprNode {
 public:
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
};

/** Sequence node. Evaluates a, then b. */
class IntSequenceNode : public IntSimpleRValueExprNode {
 public:
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
};

/** If/then/else node. */
class IntConditionalNode : public IntSimpleRValueExprNode {
 public:
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
};

/** Logical operator (And/Or/Xor). Generates code for a logical operation using
    short-circuit evaluation. A short-cut jump determines whether the result
    is already known after evaluating the first operand; if both operands have
    to be evaluated, a binary operation combines them into the real result. */
class IntLogicalNode : public IntSimpleRValueExprNode {
    uint8_t shortcut_jump;
    uint8_t binary_op;
 public:
    IntLogicalNode(uint8_t shortcut_jump, uint8_t binary_op);
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileCondition(IntBytecodeObject& bco, const IntCompilationContext& cc, IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff);
};

/** Member access. Implements "someexpr.member". */
class IntMemberNode : public IntExprNode {
    string_t name;
    IntExprNode* expr;
 public:
    IntMemberNode(string_t name);
    ~IntMemberNode();
    void setExpression(IntExprNode* expr) throw();
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs);
    void compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileWrite(IntBytecodeObject& bco, const IntCompilationContext& cc);
};

/** Identifier access. Implements a freestanding identifier ("x"). */
class IntIdentifierNode : public IntExprNode {
    string_t name;
 public:
    IntIdentifierNode(string_t name);
    ~IntIdentifierNode();
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs);
    void compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileWrite(IntBytecodeObject& bco, const IntCompilationContext& cc);
    const string_t& getIdentifier() const
        { return name; }
};

/** Function call. Base class for all sorts of invocations ("something(args)"). */
class IntFunctionCallNode : public IntRValueExprNode {
 protected:
    ptr_vector<IntExprNode> args;
 public:
    void addArgument(IntExprNode* arg);
    int getNumArgs() const
        { return args.size(); }
};

/** Indirect (=user-defined or member) function call.
    Implements general "someexpr(args)". */
class IntIndirectCallNode : public IntFunctionCallNode {
    IntExprNode* func;
 public:
    IntIndirectCallNode();
    ~IntIndirectCallNode();
    void setFunction(IntExprNode* func);
    void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileStore(IntBytecodeObject& bco, const IntCompilationContext& cc, IntExprNode* rhs);
    void compileRead(IntBytecodeObject& bco, const IntCompilationContext& cc);
    void compileWrite(IntBytecodeObject& bco, const IntCompilationContext& cc);
};



/** Set child nodes for unary expression. IntSimpleRValueExprNode will take ownership. */
inline void
IntSimpleRValueExprNode::setUnary(IntExprNode* a) throw()
{
    this->a = a;
}

/** Set child nodes for binary expression. IntSimpleRValueExprNode will take ownership. */
inline void
IntSimpleRValueExprNode::setBinary(IntExprNode* a, IntExprNode* b) throw()
{
    this->a = a;
    this->b = b;
}

/** Set child nodes for ternary expression. IntSimpleRValueExprNode will take ownership. */
inline void
IntSimpleRValueExprNode::setTernary(IntExprNode* a, IntExprNode* b, IntExprNode* c) throw()
{
    this->a = a;
    this->b = b;
    this->c = c;
}

#endif
