/**
  *  \file int/if/shipif.cc
  *  \brief Interpreter Game Interface: Ships
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "int/if/shipif.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/actions/basefix.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/mission.h"
#include "game/objtypes.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/tables.h"        // getHeadingName
#include "int/args.h"
#include "int/if/hullif.h"
#include "int/if/ifutil.h"
#include "int/if/objif.h"
#include "int/if/playerif.h"
#include "int/if/specif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "game/cspec.h"
#include "int/contextio.h"
#include "int/exec.h"
#include "int/process.h"
#include "int/propenum.h"
#include "game/cargotr.h"
#include "game/fleet.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "game/shippredict.h"
#include "util/misc.h"
#include "int/if/mapmsgif.h"

/****************************** IntShipValue *****************************/

/* @q Ship(sid:Int):Obj (Function, Context)
   Access starship properties.
   Use as
   | ForEach Ship Do ...
   or
   | With Ship(n) Do ...

   @diff This function was available for use in %With under the name %Ships() since PCC 1.0.6.
   Do not use the name %Ships in new code, it is not supported by PCC2; use %Ship instead.

   @see int:index:group:shipproperty|Ship Properties, int:index:group:shipcommand|Ship Commands
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFShipGet(IntArgBlock& args)
{
    int32_t id;

    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_SHIPS))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntShipContext(id);
}

IntContext*
IFShipMake()
{
    IntShipContext* sc = new IntShipContext(0);
    return sc->initFirst();
}

int32_t
IFShipDim(int32_t dim)
{
    return dim == 0 ? 1 : NUM_SHIPS+1;
}

/****************************** Ship Methods *****************************/

/* @q SetComment s:Str (Ship Command)
   Set ship comment.
   @see Comment (Ship Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetComment(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (IntValue* v = args.getNext()) {
        IntStringValue sv(v->toString(false));
        int sid = obj.getId();
        getShipProperties(sid).set(sp_Comment, &sv);
        univ.ty_any_ships.getObjectByIndex(sid).markDirty();
    }
}

/* @q SetFCode fc:Str (Ship Command)
   Set ship friendly code.
   @see FCode (Ship Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetFCode(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispFCode, args.getNext());
}

/* @q SetEnemy n:Int (Ship Command)
   Set ship primary enemy. %n is an integer, either 0 (=no enemy) or 1..11 (player).
   @see Enemy$ (Ship Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetEnemy(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispEnemyId, args.getNext());
}

/* @q SetSpeed sp:Int (Ship Command)
   Set ship warp speed. %sp is an integer between 0 and 9.
   @see Speed$ (Ship Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetSpeed(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispSpeedId, args.getNext());
}

/* @q SetName n:Str (Ship Command)
   Set ship name.
   @see Name (Ship Property)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetName(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispName, args.getNext());
}

/* @q SetMission m:Int, Optional i:Int, t:Int (Ship Command)
   Set ship mission.
   %m is the mission number, %i and %t are the Intercept and Tow parameters, respectively.
   @see Mission$ (Ship Property), Mission.Intercept, Mission.Tow
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetMission(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t m;
    int32_t i = 0, t = 0;

    // Evaluate arguments
    args.checkArgumentCount(1, 3);
    if (!checkIntArg(m, args.getNext(), 0, 10000))
        return;
    checkIntArg(i, args.getNext(), 0, 10000);
    checkIntArg(t, args.getNext(), 0, 10000);

    // Set mission on ship
    GShip& sh = static_cast<GShip&>(obj);
    if (sh.getShipKind() == GShip::CurrentShip) {
        if (!setFleetMission(univ, sh.getId(), m, i, t)) {
            throw GError(GError::eFleet, GError::eFleet);
        }
    } else {
        throw IntError::notAssignable();
    }
}

static void
setBaseOrder(GUniverse& univ, TShipyardAction order, GShip& sh)
{
    // Find the base we're at and verify the command
    int pid = univ.getPlanetAt(sh.getPos());
    if (GError e = checkBaseAction(order, univ, sh.getId(), pid))
        throw e;

    // Do it
    univ.getPlanet(pid).setBaseShipyardOrder(mp16_t(order), mp16_t(sh.getId()));
    sh.markDirty();
}

/* @q FixShip (Ship Command)
   Repair this ship at the starbase.
   Changes the base's order to repair this ship.
   @see FixShip (Planet Command)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipFixShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    setBaseOrder(univ, sya_Fix, static_cast<GShip&>(obj));
}

/* @q RecycleShip (Ship Command)
   Recycle this ship at the starbase.
   Changes the base's order to recycle this ship.
   @see FixShip (Planet Command)
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipRecycleShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    setBaseOrder(univ, sya_Recycle, static_cast<GShip&>(obj));
}

/* @q SetWaypoint x:Int, y:Int (Ship Command)
   Change the ship's waypoint.
   When playing on a wrapped map, this sets the waypoint to move the shortest possible way to the specified coordinates.
   @see MoveTo
   @since PCC 1.0.5, PCC2 1.99.9 */
static void
IFShipSetWaypoint(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t x, y;

    // Evaluate arguments
    args.checkArgumentCount(2);
    if (!checkIntArg(x, args.getNext(), 0, 10000))
        return;
    if (!checkIntArg(y, args.getNext(), 0, 10000))
        return;

    // Set ship waypoint
    GShip& sh = static_cast<GShip&>(obj);
    if (sh.getShipKind() == GShip::CurrentShip) {
        if (!setFleetWaypoint(univ, sh.getId(), GPoint(x, y)))
            throw GError(GError::eFleet, GError::eFleet);
    } else {
        throw IntError::notAssignable();
    }
}

static GCargoContainer*
makeShipTransfer(GShip& from, GShip& to)
{
    // Make transfer object for the "from" side
    if (from.getShipKind() == GShip::CurrentShip) {
        return new GShipTransfer(from);
    } else {
        if (to.isTransporterInUse(GShip::t_Transfer)) {
            throw IntError(GError::eFacility);
        }
        if (to.getTransporter(GShip::t_Transfer).id != 0
            && to.getTransporter(GShip::t_Transfer).id != from.getId()
            && to.isTransporterActive(GShip::t_Transfer))
        {
            if (!to.isTransporterCancellable(GShip::t_Transfer)) {
                throw IntError("Transporter jammed");
            }
            to.cancelTranporter(GShip::t_Transfer);
        }
        return new GShipTransporterTransfer(to, GShip::t_Transfer, from);
    }
}

GCargoContainer*
makePlanetTransfer(GUniverse& univ, int pid, GShip& to)
{
    if (pid == 0) {
        if (to.isTransporterInUse(GShip::t_Unload)) {
            throw IntError(GError::eFacility);
        }
        return new GShipTransporterTransfer(to);
    } else {
        GPlanet& pl = univ.getPlanet(pid);
        if (pl.isVisible() && pl.hasFullPlanetData() && pl.getOwner() == to.getOwner()) {
            // Direct transfer
            return new GPlanetTransfer(pl);
        } else {
            if (to.isTransporterInUse(GShip::t_Unload)) {
                throw IntError(GError::eFacility);
            }
            return new GShipTransporterTransfer(to, GShip::t_Unload, pl);
        }
    }
}

namespace {
    enum {
        flagOverload = 1,
        flagSup2MC   = 2,
        flagNoFail   = 4,
        flagJettison = 8
    };
}

/** Script-based cargo transfer.
    \param exc       Execution context of script
    \param transfer  Readily-configured GCargoTransfer
    \param cargospec Cargospec provided by user
    \param side      GCargoTransfer side to take away cargo from
    \param flags     Flags provided by user */
void
doScriptTransfer(IntExecutionContext& exc, GCargoTransfer& transfer, string_t cargospec, int side, int flags)
{
    // Parse cargospec
    GCargoSpec cs;
    if (!cs.parse(cargospec, false))
        throw IntError("Invalid cargo specification");

    // Configure
    if (flags & flagOverload) {
        transfer[0].setFlags(transfer[0].getFlags() | GCargoContainer::can_Overload);
        transfer[1].setFlags(transfer[1].getFlags() | GCargoContainer::can_Overload);
    }
    bool sup2mc = (flags & flagSup2MC) != 0
        && (transfer[0].hasFlag(GCargoContainer::can_Sup2mc)
            || transfer[1].hasFlag(GCargoContainer::can_Sup2mc));
    bool partial = (flags & flagNoFail) != 0;

    // Do it
    for (int i = 0; i <= el_Torps; ++i) {
        const GCargoType ct = GCargoType(i);
        if (cs.isValidType(ct)) {
            if (ct == el_Torps) {
                // This is a hack, but it's the same one as used in PCC 1.x :-)
                // Because GCargoTransfer needs an exact torpedo type, but we only
                // know that the user wants torpedoes, just try them all. Only one
                // of them will work (or none if the units are incompatible).
                for (int tt = 1; tt <= NUM_TORPS; ++tt)
                    cs.remove(ct, transfer.move(getCargoTypeFromTorpType(tt), cs.get(ct), side, partial, sup2mc));
            } else {
                cs.remove(ct, transfer.move(ct, cs.get(ct), side, partial, sup2mc));
            }
        }
    }

    // Evaluate
    if (partial) {
        IntStringValue sv(cs.toCargoSpecString());
        exc.setVariable("CARGO.REMAINDER", &sv);
    } else {
        if (!cs.isEmpty()) {
            throw IntError("Cargo transfer failed");
        }
    }
}

/* @q CargoTransfer amount:Cargo, target:Int, Optional flags:Any (Ship Command, Planet Command)
   Transfers cargo to a ship.
   %target specifies the target ship Id,
   %cargo is the amount and type to transfer, as a {int:index:type:cargo|cargo set}.
   It may contain negative values to beam cargo back from the target if rules allow.

   The optional third parameter, %options, is a string containing some options:
   - <tt>"O"</tt>: permit overloading the ships.
   - <tt>"S"</tt>: sell supplies on-the-fly (not possible for the ship/ship version of this command).
   - <tt>"N"</tt>: don't generate an error when the cargo can not completely be transferred.
     Instead, sets the variable <tt>{Cargo.Remainder}</tt> accordingly.

   For example,
   | CargoTransfer 10, "300n"
   transfers 300 kt Neutronium to ship 10.
   It will fail with an error if the current ship doesn't have that much,
   or there is not enough space in ship 10's fuel tank.
   If you use
   | CargoTransfer 10, "300n", "n"
   PCC will transfer as much as possible.
   Assuming that the current ship only has 20 kt fuel (and ship 10 has enough space),
   this will set {Cargo.Remainder} to "280N", because that's the amount that was not transferred.
   To test for a successful transfer, you can then use
   | If Not Cargo.Remainder Then Print "Successful."

   Though you seem to be able to call this command beam cargo off foreign ships,
   this can only be used to cancel pending transfer orders.
   That is, if ship 355 is a foreign one and ship 10 is yours: for
   | With Ship(355) Do CargoTransfer 10, "10n"
   to succeed, you must have transferred 10 kt Neutronium there using the ship/enemy-ship transporter,
   e.g. with the command
   | With Ship(10) Do CargoTransfer 355, "10n"
   The game rules do not allow asking foreign ships for stuff; the other ship must explicitly send it.
   @see CargoUnload, CargoUpload, CargoTransferWait
   @since PCC 1.0.10, PCC2 1.99.12 */
static void
IFShipCargoTransfer(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    // Args are cargospec, target ship Id, flags O/S/N (overload, supply sale, no-fail)
    string_t cargospec;
    int32_t target_sid;
    int32_t flags = 0;
    args.checkArgumentCount(2, 3);
    if (!checkStringArg(cargospec, args.getNext()))
        return;
    if (!checkIntArg(target_sid, args.getNext(), 1, NUM_SHIPS))
        return;
    checkFlagArg(flags, 0, args.getNext(), "OSN");

    // Validate
    GShip& sh       = static_cast<GShip&>(obj);
    GShip& target   = univ.getShip(target_sid);
    if (!target.isVisible() || !sh.isVisible()) {
        throw IntError("Ship does not exist");
    }
    if (target.getPos() != sh.getPos()) {
        throw IntError(GError::ePos);
    }
    if (target.getId() == sh.getId()) {
        // Forbid transfer from one ship to itself. PCC 1.x silently ignores this,
        // leading to subtle effects with Cargo.Remainder. So better fail it.
        // It's useless anyway.
        throw IntError(GError::ePerm);
    }
    if (target.getShipKind() != GShip::CurrentShip && sh.getShipKind() != GShip::CurrentShip) {
        // Neither is a current ship.
        throw IntError(GError::eNotOwner);
    }

    // Build the transfer structures
    GCargoTransfer tx;
    tx.addNewContainer(makeShipTransfer(sh, target));
    tx.addNewContainer(makeShipTransfer(target, sh));
    doScriptTransfer(exc, tx, cargospec, 0, flags);
    tx.doCommit();
}

static void
doShipCargoUnUpload(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args, int side)
{
    // Args are cargospec, flags O/S/N/J (overload, supply sale, no-fail, jettison permission) + proxy Id
    // FIXME: proxy transfer support
    string_t cargospec;
    int32_t flags = 0;
    args.checkArgumentCount(1, 2);
    if (!checkStringArg(cargospec, args.getNext()))
        return;
    checkFlagArg(flags, 0, args.getNext(), "OSNJ");

    // Validate
    GShip& sh = static_cast<GShip&>(obj);
    if (!sh.isVisible()) {
        throw IntError("Ship does not exist");
    }
    if (sh.getShipKind() != GShip::CurrentShip) {
        throw IntError(GError::eNotOwner);
    }
    int pid = sh.getOrbitPlanetId();
    if (pid == 0 && (flags & flagJettison) == 0) {
        throw IntError("Not at planet");
    }

    // Build the transfer structures
    GCargoTransfer tx;
    tx.addNewContainer(new GShipTransfer(sh));
    tx.addNewContainer(makePlanetTransfer(univ, pid, sh));
    doScriptTransfer(exc, tx, cargospec, side, flags);
    tx.doCommit();
}

/* @q CargoUnload amount:Cargo, Optional flags:Str (Ship Command)
   Unload cargo to planet (or jettison).
   %cargo is the amount and type to transfer, as a {int:index:type:cargo|cargo set}.

   The optional second parameter, %options, is a string containing some options:
   - <tt>"O"</tt>: permit overloading the ships.
   - <tt>"S"</tt>: sell supplies on-the-fly.
   - <tt>"N"</tt>: don't generate an error when the cargo can not completely be transferred.
     Instead, sets the variable <tt>{Cargo.Remainder}</tt> accordingly.
   - <tt>"J"</tt>: permit jettison. By default, this command will fail if the ship does not orbit a planet.

   This command is equivalent to
   | CargoUpload CMul(amount, -1), flags
   @see CargoUpload, CargoTransfer
   @since PCC 1.0.10, PCC2 1.99.12 */
static void
IFShipCargoUnload(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    doShipCargoUnUpload(obj, univ, exc, args, 0);
}

/* @q CargoUpload amount:Cargo, Optional flags:Str (Ship Command)
   Load cargo from planet.
   %amount is the amount and type to transfer, as a {int:index:type:cargo|cargo set}.

   The optional second parameter, %options, is a string containing some options:
   - <tt>"O"</tt>: permit overloading the ships.
   - <tt>"S"</tt>: sell supplies on-the-fly.
   - <tt>"N"</tt>: don't generate an error when the cargo can not completely be transferred.
     Instead, sets the variable <tt>{Cargo.Remainder}</tt> accordingly.
   - <tt>"J"</tt>: permit jettison. By default, this command will fail if the ship does not orbit a planet.
     Use this option if you want to use %CargoUpload to get back cargo you jettisoned earlier this turn.
   @see CargoUnload, CargoTransfer, CargoUploadWait
   @since PCC 1.0.10, PCC2 1.99.12 */
static void
IFShipCargoUpload(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    doShipCargoUnUpload(obj, univ, exc, args, 1);
}

/* @q SetFleet fid:Int (Ship Command)
   Sets the fleet this ship is member of.
   %fid can be one of the following:
   - 0 - the ship is not member of any fleet.
   - the ship's id - the ship is in a fleet alone.
   - a fleet id - the ship becomes member of the specified fleet.

   If you're putting the ship into a fleet, but it is already member of a different one,
   it leaves its old fleet first (as if you had written <tt>SetFleet 0</tt>).
   This makes a difference when you're putting a fleet leader into another fleet.
   @see Fleet$ (Ship Property) */
static void
IFShipSetFleet(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setShipProperty(static_cast<GShip&>(obj), univ, ispFleetId, args.getNext());
}

/*********************** IntShipArrayProperty **********************/

namespace {
    class IntShipArrayProperty : public IntIndexableValue {
     public:
        IntShipArrayProperty(const GShip& ship, IntShipProperty property);

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        int32_t getDimension(int32_t dim);
        IntContext* makeFirstContext();

        // IntValue:
        IntShipArrayProperty* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);

     private:
        const GShip& ship;
        const IntShipProperty property;
    };
}

IntShipArrayProperty::IntShipArrayProperty(const GShip& ship, const IntShipProperty property)
    : IntIndexableValue(),
      ship(ship),
      property(property)
{ }

IntValue*
IntShipArrayProperty::get(IntArgBlock& args)
{
    switch (property) {
     case ispScore:
        /* @q Score:Int() (Ship Property, Planet Property)
           Get unit's score of a given type.

           PHost can associate various scores with ships and planets (utilX.dat records 49 and 50).
           This property makes these scores available to scripts.
           Valid parameters can be found in the PHost documentation.
           As of PHost 4.1, the following values are valid:

           - Score(1): experience level (same as {Level}).
           - Score(2): experience points.

           This property yields EMPTY if the respective score does not exist or is not known.
           @since PCC2 1.99.21, PCC 1.1.16 */
        {
            args.checkArgumentCount(1);
            int32_t id;
            if (!checkIntArg(id, args.getNext(), 0, 0x7FFF)) {
                return 0;
            }
            int32_t score = ship.unit_scores.getScore(ship_score_definitions.lookupScore(id));
            if (score < 0) {
                return 0;
            } else {
                return makeIntValue(score);
            }
        }

     case ispHasFunction:
        /* @q HasFunction:Bool() (Ship Property)
           True if the ship has the specified hull function used as index.
           The index is either the name ("Gravitonic") or number (7) of the function,
           as it can be used in SHIPLIST.TXT.

           This property considers all functions assigned to this ship type as well as to this individual ship,
           and honors level restrictions.
           @since PCC2 1.99.21, PCC 1.1.15 */
        {
            args.checkArgumentCount(1);
            string_t func;
            if (!checkStringArg(func, args.getNext())) {
                return 0;
            }
            int32_t funcId;
            if (stringValue(func, funcId)) {
                // accept
            } else if (const GBasicHullFunction* hf = GHullFunctionData::instance.getBasicFunctionByName(func, false)) {
                funcId = hf->getId();
            } else {
                throw IntError("Invalid hull function name");
            }
            return makeBoolValue(ship.canDoSpecial(funcId));
        }

     default:
        return 0;
    }
}

void
IntShipArrayProperty::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

int32_t
IntShipArrayProperty::getDimension(int32_t /*dim*/)
{
    // These properties behave like hashes, not arrays, so don't claim to be an array
    return 0;
}

IntContext*
IntShipArrayProperty::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

IntShipArrayProperty*
IntShipArrayProperty::clone()
{
    return new IntShipArrayProperty(ship, property);
}

string_t
IntShipArrayProperty::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntShipArrayProperty::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

/***************************** IntShipContext ****************************/

namespace {
    enum { ShipProperty, HullProperty, ComponentProperty, OwnerProperty, EnemyProperty, ShipMethod };
}

static const TIntMapping ship_mapping[] = {
    { "AUX",                       ispAuxName,                   ShipProperty,      thString },
    { "AUX$",                      ispAuxId,                     ShipProperty,      thInt },
    { "AUX.AMMO",                  ispAuxAmmo,                   ShipProperty,      thInt },
    { "AUX.COUNT",                 ispAuxCount,                  ShipProperty,      thInt },
    { "AUX.SHORT",                 ispAuxShort,                  ShipProperty,      thString },
    { "BEAM",                      ispBeamName,                  ShipProperty,      thString },
    { "BEAM$",                     ispBeamId,                    ShipProperty,      thInt },
    { "BEAM.COUNT",                ispBeamCount,                 ShipProperty,      thInt },
    { "BEAM.MAX",                  ihpMaxBeams,                  HullProperty,      thInt },
    { "BEAM.SHORT",                ispBeamShort,                 ShipProperty,      thString },
    { "CARGO.COLONISTS",           ispCargoColonists,            ShipProperty,      thInt },
    { "CARGO.D",                   ispCargoD,                    ShipProperty,      thInt },
    { "CARGO.FREE",                ispCargoFree,                 ShipProperty,      thInt },
    { "CARGO.M",                   ispCargoM,                    ShipProperty,      thInt },
    { "CARGO.MAX",                 ihpMaxCargo,                  HullProperty,      thInt },
    { "CARGO.MAXFUEL",             ihpMaxFuel,                   HullProperty,      thInt },
    { "CARGO.MONEY",               ispCargoMoney,                ShipProperty,      thInt },
    { "CARGO.N",                   ispCargoN,                    ShipProperty,      thInt },
    { "CARGO.STR",                 ispCargoStr,                  ShipProperty,      thString },
    { "CARGO.SUPPLIES",            ispCargoSupplies,             ShipProperty,      thInt },
    { "CARGO.T",                   ispCargoT,                    ShipProperty,      thInt },
    { "CARGOTRANSFER",             ismCargoTransfer,             ShipMethod,        thProcedure},
    { "CARGOUNLOAD",               ismCargoUnload,               ShipMethod,        thProcedure },
    { "CARGOUPLOAD",               ismCargoUpload,               ShipMethod,        thProcedure },
    { "CREW",                      ispCrew,                      ShipProperty,      thInt },
    { "CREW.NORMAL",               ihpMaxCrew,                   HullProperty,      thInt },
    { "DAMAGE",                    ispDamage,                    ShipProperty,      thInt },
    { "ENEMY",                     iplShortName,                 EnemyProperty,     thString },
    { "ENEMY$",                    ispEnemyId,                   ShipProperty,      thInt },
    { "ENEMY.ADJ",                 iplAdjName,                   EnemyProperty,     thString },
    { "ENGINE",                    ispEngineName,                ShipProperty,      thString },
    { "ENGINE$",                   ispEngineId,                  ShipProperty,      thInt },
    { "ENGINE.COUNT",              ihpNumEngines,                HullProperty,      thInt },
    { "FCODE",                     ispFCode,                     ShipProperty,      thString },
    { "FIGHTER.BAYS",              ispFighterBays,               ShipProperty,      thInt },
    { "FIGHTER.COUNT",             ispFighterCount,              ShipProperty,      thInt },
    { "FIXSHIP",                   ismFixShip,                   ShipMethod,        thProcedure },
    { "FLEET",                     ispFleet,                     ShipProperty,      thString },
    { "FLEET$",                    ispFleetId,                   ShipProperty,      thInt },
    { "FLEET.NAME",                ispFleetName,                 ShipProperty,      thString },
    { "FLEET.STATUS",              ispFleetStatus,               ShipProperty,      thString },
    { "HASFUNCTION",               ispHasFunction,               ShipProperty,      thArray },
    { "HEADING",                   ispHeadingName,               ShipProperty,      thString },
    { "HEADING$",                  ispHeadingAngle,              ShipProperty,      thInt },
    { "HULL",                      icpName,                      ComponentProperty, thString },
    { "HULL$",                     icpId,                        ComponentProperty, thInt },
    { "HULL.SHORT",                icpNameShort,                 ComponentProperty, thString },
    { "HULL.SPECIAL",              ispHullSpecial,               ShipProperty,      thString },
    { "ID",                        ispId,                        ShipProperty,      thInt },
    { "LEVEL",                     ispLevel,                     ShipProperty,      thInt },
    { "LOC",                       ispLoc,                       ShipProperty,      thString },
    { "LOC.X",                     ispLocX,                      ShipProperty,      thInt },
    { "LOC.Y",                     ispLocY,                      ShipProperty,      thInt },
    { "MARK",                      ismMark,                      ShipMethod,        thProcedure },
    { "MARKED",                    ispMarked,                    ShipProperty,      thBool },
    { "MASS",                      ispMass,                      ShipProperty,      thInt },
    { "MESSAGES",                  ispMessages,                  ShipProperty,      thArray },
    { "MISSION",                   ispMissionName,               ShipProperty,      thString },
    { "MISSION$",                  ispMissionId,                 ShipProperty,      thInt },
    { "MISSION.INTERCEPT",         ispMissionIntercept,          ShipProperty,      thInt },
    { "MISSION.SHORT",             ispMissionShort,              ShipProperty,      thString },
    { "MISSION.TOW",               ispMissionTow,                ShipProperty,      thInt },
    { "MOVE.ETA",                  ispMoveETA,                   ShipProperty,      thInt },
    { "MOVE.FUEL",                 ispMoveFuel,                  ShipProperty,      thInt },
    { "NAME",                      ispName,                      ShipProperty,      thString },
    { "ORBIT",                     ispOrbitName,                 ShipProperty,      thString },
    { "ORBIT$",                    ispOrbitId,                   ShipProperty,      thInt },
    { "OWNER",                     iplShortName,                 OwnerProperty,     thString },
    { "OWNER$",                    iplId,                        OwnerProperty,     thInt },
    { "OWNER.ADJ",                 iplAdjName,                   OwnerProperty,     thString },
    { "OWNER.REAL",                ispRealOwner,                 ShipProperty,      thInt },
    { "PLAYED",                    ispPlayed,                    ShipProperty,      thBool },
    { "RECYCLESHIP",               ismRecycleShip,               ShipMethod,        thProcedure },
    { "SCORE",                     ispScore,                     ShipProperty,      thArray },
    { "SETCOMMENT",                ismSetComment,                ShipMethod,        thProcedure },
    { "SETENEMY",                  ismSetEnemy,                  ShipMethod,        thProcedure },
    { "SETFCODE",                  ismSetFCode,                  ShipMethod,        thProcedure },
    { "SETFLEET",                  ismSetFleet,                  ShipMethod,        thProcedure },
    { "SETMISSION",                ismSetMission,                ShipMethod,        thProcedure },
    { "SETNAME",                   ismSetName,                   ShipMethod,        thProcedure },
    { "SETSPEED",                  ismSetSpeed,                  ShipMethod,        thProcedure },
    { "SETWAYPOINT",               ismSetWaypoint,               ShipMethod,        thProcedure },
    { "SPEED",                     ispSpeedName,                 ShipProperty,      thString },
    { "SPEED$",                    ispSpeedId,                   ShipProperty,      thInt },
    { "TASK",                      ispTask,                      ShipProperty,      thBool },
    { "TECH.HULL",                 icpTech,                      ComponentProperty, thInt },
    { "TORP",                      ispTorpName,                  ShipProperty,      thString },
    { "TORP$",                     ispTorpId,                    ShipProperty,      thInt },
    { "TORP.COUNT",                ispTorpCount,                 ShipProperty,      thInt },
    { "TORP.LCOUNT",               ispTorpLCount,                ShipProperty,      thInt },
    { "TORP.LMAX",                 ihpMaxTorpLaunchers,          HullProperty,      thInt },
    { "TORP.SHORT",                ispTorpShort,                 ShipProperty,      thString },
    { "TRANSFER.SHIP",             ispTransferShip,              ShipProperty,      thBool },
    { "TRANSFER.SHIP.COLONISTS",   ispTransferShipColonists,     ShipProperty,      thInt },
    { "TRANSFER.SHIP.D",           ispTransferShipD,             ShipProperty,      thInt },
    { "TRANSFER.SHIP.ID",          ispTransferShipId,            ShipProperty,      thInt },
    { "TRANSFER.SHIP.M",           ispTransferShipM,             ShipProperty,      thInt },
    { "TRANSFER.SHIP.N",           ispTransferShipN,             ShipProperty,      thInt },
    { "TRANSFER.SHIP.NAME",        ispTransferShipName,          ShipProperty,      thInt },
    { "TRANSFER.SHIP.SUPPLIES",    ispTransferShipSupplies,      ShipProperty,      thInt },
    { "TRANSFER.SHIP.T",           ispTransferShipT,             ShipProperty,      thInt },
    { "TRANSFER.UNLOAD",           ispTransferUnload,            ShipProperty,      thBool },
    { "TRANSFER.UNLOAD.COLONISTS", ispTransferUnloadColonists,   ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.D",         ispTransferUnloadD,           ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.ID",        ispTransferUnloadId,          ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.M",         ispTransferUnloadM,           ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.N",         ispTransferUnloadN,           ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.NAME",      ispTransferUnloadName,        ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.SUPPLIES",  ispTransferUnloadSupplies,    ShipProperty,      thInt },
    { "TRANSFER.UNLOAD.T",         ispTransferUnloadT,           ShipProperty,      thInt },
    { "TYPE",                      ispTypeStr,                   ShipProperty,      thString },
    { "TYPE.SHORT",                ispTypeChar,                  ShipProperty,      thString },
    { "UNMARK",                    ismUnmark,                    ShipMethod,        thProcedure },
    { "WAYPOINT",                  ispWaypointName,              ShipProperty,      thString },
    { "WAYPOINT.DIST",             ispWaypointDistance,          ShipProperty,      thFloat },
    { "WAYPOINT.DX",               ispWaypointDX,                ShipProperty,      thInt },
    { "WAYPOINT.DY",               ispWaypointDY,                ShipProperty,      thInt },
    { "WAYPOINT.PLANET",           ispWaypointPlanetId,          ShipProperty,      thInt },
    { "WAYPOINT.X",                ispWaypointX,                 ShipProperty,      thInt },
    { "WAYPOINT.Y",                ispWaypointY,                 ShipProperty,      thInt },
};

static const IntObjectProcedureValue::call_t ship_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
    IFShipSetComment,           // 2
    IFShipSetFCode,             // 3
    IFShipSetEnemy,             // 4
    IFShipSetSpeed,             // 5
    IFShipSetName,              // 6
    IFShipSetMission,           // 7
    IFShipFixShip,              // 8
    IFShipRecycleShip,          // 9
    IFShipSetWaypoint,          // 10
    IFShipCargoTransfer,        // 11
    IFShipCargoUnload,          // 12
    IFShipCargoUpload,          // 13
    IFShipSetFleet,             // 14
};

IntShipContext::IntShipContext(int sid)
    : IntObjectContext(current_anyship_type, sid)
{ }

IntShipContext::~IntShipContext()
{ }

static int32_t
lookupShipProperty(const IntNameQuery& name)
{
    static const int32_t nproperties = countof(ship_mapping);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = ship_property_names.lookup(name);
    if (ix != ship_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, ship_mapping, nproperties);
}

int32_t
IntShipContext::lookup(const IntNameQuery& name)
{
    /* Check prefix */
    if (name.startsWith("SHIP.", 5)) {
        return lookupShipProperty(IntNameQuery(name, 5));
    } else {
        return lookupShipProperty(name);
    }
}

void
IntShipContext::set(int32_t index, IntValue* value)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip* sh = dynamic_cast<GShip*>(getObject());
        if (sh == 0) {
            throw IntError::notAssignable();
        }
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            setShipProperty(*sh, *getCurrentUniverse(), IntShipProperty(ship_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        /* User property */
        getShipProperties(getCurrentIndex()).set(index - countof(ship_mapping), value);
        if (GObject* obj = getObject()) {
            obj->markDirty();
        }
    }
}

IntValue*
IntShipContext::get(int32_t index)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip* sh = dynamic_cast<GShip*>(getObject());
        if (sh == 0) {
            return 0;
        }
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            return getShipProperty(*sh, *getCurrentUniverse(), IntShipProperty(ship_mapping[index].index));
         case HullProperty:
            if (sh->getHullId().isKnown())
                return getHullProperty(sh->getHull(), IntHullProperty(ship_mapping[index].index));
            else
                return 0;
         case ComponentProperty:
            if (sh->getHullId().isKnown())
                return getComponentProperty(sh->getHull(), IntComponentProperty(ship_mapping[index].index));
            else
                return 0;
         case OwnerProperty:
            return getPlayerProperty(sh->getOwner(), IntPlayerProperty(ship_mapping[index].index));
         case EnemyProperty:
            if (sh->getPrimaryEnemy().isKnown())
                return getPlayerProperty(sh->getPrimaryEnemy(), IntPlayerProperty(ship_mapping[index].index));
            else
                return 0;
         case ShipMethod:
            return new IntObjectProcedureValue(*sh, *getCurrentUniverse(), ship_methods[ship_mapping[index].index]);
         default:
            return 0;
        }
    } else {
        /* User property */
        return IntValue::cloneOf(getShipProperties(getCurrentIndex())[index - countof(ship_mapping)]);
    }
}

IntContext*
IntShipContext::clone()
{
    return new IntShipContext(getCurrentIndex());
}

string_t
IntShipContext::toString(bool /*readable*/)
{
    return format("Ship(%d)", getCurrentIndex());
}

void
IntShipContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Ship;
    sv.value = getCurrentIndex();
}

void
IntShipContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumNames(ship_property_names);
    acceptor.enumMapping(ship_mapping, countof(ship_mapping));
}

GObject*
IntShipContext::getObject()
{
    if (current_histship_type.isValidIndex(getCurrentIndex())) {
        return &current_histship_type.getObjectByIndex(getCurrentIndex());
    } else {
        return 0;
    }
}

/** Classify ship. This yields the ship's category as a string, 0 if unknown. */
static const char*
classifyShip(const GShip& sh)
{
    int beams = sh.getNumBeams().isKnown()         ? sh.getNumBeams()         : sh.getHullId().isKnown() ? sh.getHull().getMaxBeams()     : -1;
    int tubes = sh.getNumTorpLaunchers().isKnown() ? sh.getNumTorpLaunchers() : sh.getHullId().isKnown() ? sh.getHull().getMaxLaunchers() : -1;
    int bays  = sh.getNumBays().isKnown()          ? sh.getNumBays()          : sh.getHullId().isKnown() ? sh.getHull().getNumBays()      : -1;

    if (bays > 0)
        return "Carrier";
    else if (tubes > 0)
        return "Torpedo Ship";
    else if (beams > 0)
        return "Beam Weapons";
    else if (tubes == 0 && beams == 0 && bays == 0)
        return "Freighter";
    else
        return 0;
}

/** Get ship property.
    \param sh Ship to query
    \param isp Property to query */
IntValue*
getShipProperty(const GShip& sh, GUniverse& univ, const IntShipProperty isp)
{
    /* Combat participant properties often share names and meaning with ship properties,
       and are therefore documented here as well for brevity. Documenting them separately
       in getVcrSideProperty() would generate new documentation fragments and disambiguations
       for every item. */
    switch (isp) {
     case ispAuxId:
        /* @q Aux$:Int (Ship Property, Combat Participant Property)
           Type of secondary weapon.
           - 1..10 for torpedoes
           - 11 for fighters
           - EMPTY if no secondary weapon, or not known. */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeIntValue(11);
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeIntValue(sh.getTorpType());
        else
            return 0;
     case ispAuxAmmo:
        /* @q Aux.Ammo:Int (Ship Property, Combat Participant Property)
           Number of fighters/torpedoes. */
        return makeIntFromMaybe(sh.getAmmoRaw());
     case ispAuxCount:
        /* @q Aux.Count:Int (Ship Property, Combat Participant Property)
           Number of fighter bays/torpedo launchers. */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeIntValue(sh.getNumBays());
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0 && sh.getNumTorpLaunchers().isKnown())
            return makeIntValue(sh.getNumTorpLaunchers());
        else
            return 0;
     case ispAuxShort:
        /* @q Aux.Short:Str (Ship Property, Combat Participant Property)
           Secondary weapon type, short name.
           @see Aux (Ship Property) */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeStringValue("Ftr");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispAuxName:
        /* @q Aux:Str (Ship Property, Combat Participant Property)
           Secondary weapon type, full name.
           Either a torpedo system name, "Fighters", or EMPTY. */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0)
            return makeStringValue("Fighters");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispBeamId:
        /* @q Beam$:Int (Ship Property, Combat Participant Property)
           Beam type. 0 if none, EMPTY if not known. */
        return makeIntFromMaybe(sh.getBeamType());
     case ispBeamCount:
        /* @q Beam.Count:Int (Ship Property, Combat Participant Property)
           Number of beams. */
        return makeIntFromMaybe(sh.getNumBeams());
     case ispBeamShort:
        /* @q Beam.Short:Str (Ship Property, Combat Participant Property)
           Beam type, short name. */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getShortName());
        else
            return 0;
     case ispBeamName:
        /* @q Beam:Str (Ship Property, Combat Participant Property)
           Beam type, full name. */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getName());
        else
            return 0;
     case ispCargoColonists:
        /* @q Cargo.Colonists:Int (Ship Property)
           Number of colonists aboard this ship. */
        return makeIntFromMaybe(sh.getCargoRaw(el_Colonists));
     case ispCargoD:
        /* @q Cargo.D:Int (Ship Property)
           Duranium aboard this ship, kilotons. */
        return makeIntFromMaybe(sh.getCargoRaw(el_Duranium));
     case ispCargoFree:
        /* @q Cargo.Free:Int (Ship Property)
           Free cargo room. */
        return makeIntFromMaybe(sh.getFreeCargo());
     case ispCargoM:
        /* @q Cargo.M:Int (Ship Property)
           Molybdenum aboard this ship, kilotons. */
        return makeIntFromMaybe(sh.getCargoRaw(el_Molybdenum));
     case ispCargoMoney:
        /* @q Cargo.Money:Int (Ship Property)
           Money aboard this ship. */
        return makeIntFromMaybe(sh.getCargoRaw(el_Money));
     case ispCargoN:
        /* @q Cargo.N:Int (Ship Property)
           Neutronium aboard this ship, kilotons. */
       return makeIntFromMaybe(sh.getCargoRaw(el_Neutronium));
     case ispCargoStr:
        /* @q Cargo.Str:Cargo (Ship Property)
           Cargo aboard this ship.
           String containing amounts of minerals, supplies, colonists, cash, and torpedoes/fighters. */
        {
            GCargoSpec sp;
            bool hadAny = false;
            for (GCargoType i = el_Neutronium; i <= el_Money; ++i) {
                mp16_t n = sh.getCargoRaw(i);
                if (n.isKnown()) {
                    sp.set(i, n);
                    hadAny = true;
                }
            }
            if (sh.getTorpType().isKnown() && sh.getTorpType() > 0) {
                mp16_t n = sh.getAmmoRaw();
                if (n.isKnown()) {
                    sp.set(el_Torps, n);
                }
            }

            if (hadAny) {
                return makeStringValue(sp.toCargoSpecString());
            } else {
                return 0;
            }
        }
     case ispCargoSupplies:
        /* @q Cargo.Supplies:Int (Ship Property)
           Supplies aboard this ship, kilotons. */
        return makeIntFromMaybe(sh.getCargoRaw(el_Supplies));
     case ispCargoT:
        /* @q Cargo.T:Int (Ship Property)
           Tritanium aboard this ship, kilotons. */
        return makeIntFromMaybe(sh.getCargoRaw(el_Tritanium));
     case ispCrew:
        /* @q Crew:Int (Ship Property)
           Current crew size. */
        return makeIntFromMaybe(sh.getCrew());
     case ispDamage:
        /* @q Damage:Int (Ship Property, Combat Participant Property)
           Damage level in percent. */
        return makeIntFromMaybe(sh.getDamage());
     case ispEnemyId:
        /* @q Enemy$:Int (Ship Property)
           Primary Enemy. 0=none, or a player number.
           @assignable
           @see SetEnemy (Ship Command) */
        return makeIntFromMaybe(sh.getPrimaryEnemy());
     case ispEngineId:
        /* @q Engine$:Int (Ship Property)
           Type of engine. */
        return makeIntFromMaybe(sh.getEngineType());
     case ispEngineName:
        /* @q Engine:Str (Ship Property)
           Type of engine, full name. */
        if (sh.getEngineType().isKnown())
            return makeStringValue(::getEngine(sh.getEngineType()).getName());
        else
            return 0;
     case ispFCode:
        /* @q FCode:Str (Ship Property)
           Friendly code.
           @assignable
           @see SetFCode (Ship Command) */
        if (sh.isFCodeKnown())
            return makeStringValue(sh.getFCode());
        else
            return 0;
     case ispFighterBays:
        /* @q Fighter.Bays:Int (Ship Property, Combat Participant Property)
           Number of fighter bays. */
        return makeIntFromMaybe(sh.getNumBays());
     case ispFighterCount:
        /* @q Fighter.Count:Int (Ship Property, Combat Participant Property)
           Number of fighters. */
        if (sh.getNumBays().isKnown() && sh.getNumBays() > 0 && sh.getAmmoRaw().isKnown())
            return makeIntValue(sh.getAmmoRaw());
        else
            return 0;
     case ispFleetId:
        /* @q Fleet$:Int (Ship Property)
           Id of fleet this ship is in.
           @assignable
           @see SetFleet (Ship Command) */
        return makeIntValue(sh.getFleetNumber());
     case ispFleetName:
        /* @q Fleet.Name:Str (Ship Property)
           Name of fleet this ship is leader of.
           Has a value, and is assignable, only for ships that actually are fleet leaders
           (i.e. <tt>Fleet$ = Id$</tt>).
           @assignable */
        return makeStringValue(sh.getFleetName());
     case ispFleetStatus:
        /* @q Fleet.Status:Str (Ship Property)
           Fleet status. One of
           - "leader"
           - "member"
           - "-" */
        if (int fid = sh.getFleetNumber()) {
            if (fid == sh.getId()) {
                return makeStringValue("leader");
            } else {
                return makeStringValue("member");
            }
        } else {
            return makeStringValue("-");
        }
     case ispFleet:
        /* @q Fleet:Str (Ship Property)
           Name of fleet this ship is in.
           If this ship is leader of a fleet, and the fleet has a name ({Fleet.Name}), returns that.
           Otherwise, returns the name ({Name (Ship Property)|Name}) of the leader.
           If the ship is not member of a fleet, this property is EMPTY. */
        if (int fid = sh.getFleetNumber()) {
            GShip& leader = univ.getShip(fid);
            string_t result = leader.getFleetName();
            if (result.empty()) {
                result = leader.getName(GObject::LongName);
            }
            return makeStringValue(result);
        } else {
            return 0;
        }
     case ispHeadingAngle:
        /* @q Heading$:Int (Ship Property)
           Current angle of movement, in degrees.
           EMPTY if the ship is not moving, or the angle is not known. */
        return makeIntFromMaybe(sh.getHeading());
     case ispHeadingName:
        /* @q Heading:Str (Ship Property)
           Current angle of movement, as compass direction. */
        /* heading name */
        if (sh.getHeading().isKnown())
            return makeStringValue(getHeadingName(sh.getHeading()));
        else
            return 0;
     case ispHullSpecial:
        /* @q Hull.Special:Str (Ship Property)
           Special function summary.
           This is a string identifying the major special functions of this ship.
           The string will contain each letter if and only if the ship
           has the respective ability assigned for all players.
           - "C" (Cloak, including Advanced and Hardened Cloak)
           - "H" (Hyperdrive)
           - "G" (Gravitonic accelerator)
           - "B" (Bioscan, including Full Bioscan)
           - "A" (Alchemy, including Neutronic/Aries Refinery) */
        {
            string_t result;
            if (sh.isVisible()) {
                if (sh.canDoSpecial(hf_Cloak) || sh.canDoSpecial(hf_AdvancedCloak) || sh.canDoSpecial(hf_HardenedCloak))
                    result += "C";
                if (sh.canDoSpecial(hf_Hyperdrive))
                    result += "H";
                if (sh.canDoSpecial(hf_Gravitonic))
                    result += "G";
                if (sh.canDoSpecial(hf_Bioscan) || sh.canDoSpecial(hf_FullBioscan))
                    result += "B";
                if (sh.canDoSpecial(hf_MerlinAlchemy) || sh.canDoSpecial(hf_AriesRefinery) || sh.canDoSpecial(hf_NeutronicRefinery))
                    result += "A";
            }
            return makeStringValue(result);
        }
     case ispId:
        /* @q Id:Int (Ship Property)
           Ship Id. */
        return makeIntValue(sh.getId());
     case ispLevel:
        /* @q Level:Int (Ship Property)
           Ship's experience level.
           If the experience system is not enabled, or the level is not known, yields EMPTY. */
        {
            int16_t level = sh.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ispLocX:
        /* @q Loc.X:Int (Ship Property)
           X location of ship. */
        if (sh.isVisible())
            return makeIntValue(sh.getPos().x);
        else
            return 0;
     case ispLocY:
        /* @q Loc.Y:Int (Ship Property)
           Y location of ship. */
        if (sh.isVisible())
            return makeIntValue(sh.getPos().y);
        else
            return 0;
     case ispLoc:
        /* @q Loc:Str (Ship Property)
           Location of ship, as a human-readable string.
           If the ship is at a planet, returns that planet's name and Id.
           In deep space, returns an (X,Y) pair. */
        if (sh.isVisible())
            return makeStringValue(univ.getLocationName(sh.getPos(), 0));
        else
            return 0;
     case ispMarked:
        /* @q Marked:Bool (Ship Property)
           True if ship is marked. */
        return makeBoolValue(sh.isMarked());
     case ispMass:
        /* @q Mass:Int (Ship Property)
           Mass of ship (hull, components, and cargo). */
        return makeIntFromMaybe(sh.getMass());
     case ispMessages:
        /* @q Messages:Obj() (Ship Property)
           If this ship has any messages, this property is non-null and contains an array of messages.
           Individual messages have the same form as the inbox messages (InMsg()).
           @see int:index:group:incomingmessageproperty|Incoming Message Properties
           @since PCC2 2.0.3 */
        return IntMappedMessageValue::create(sh.getAssociatedMessages());
     case ispMissionId:
        /* @q Mission$:Int (Ship Property)
           Mission number.
           @assignable
           @see SetMission (Ship Command) */
        return makeIntFromMaybe(sh.getMission());
     case ispMissionIntercept:
        /* @q Mission.Intercept:Int (Ship Property)
           Mission "Intercept" parameter.
           @assignable
           @see SetMission (Ship Command) */
        return makeIntFromMaybe(sh.getInterceptId());
     case ispMissionShort:
        /* @q Mission.Short:Str (Ship Property)
           Mission, short name. */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByShip(sh);
            if (m != 0) {
                return makeStringValue(m->getShortName());
            } else {
                return makeStringValue(format("MIT %d", int(sh.getMission())));
            }
        } else {
            return 0;
        }
     case ispMissionTow:
        /* @q Mission.Tow:Int (Ship Property)
           Mission "Tow" parameter.
           @assignable
           @see SetMission (Ship Command) */
        return makeIntFromMaybe(sh.getTowId());
     case ispMissionName:
        /* @q Mission:Str (Ship Property)
           Mission, full name. */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByShip(sh);
            if (m != 0) {
                return makeStringValue(m->getName());
            } else if (sh.getInterceptId().isKnown() && sh.getTowId().isKnown()) {
                return makeStringValue(format("M.I.T. %d (%d,%d)",
                                              int(sh.getMission()),
                                              int(sh.getInterceptId()),
                                              int(sh.getTowId())));
            } else {
                return makeStringValue(format("M.I.T. %d", int(sh.getMission())));
            }
        } else {
            return 0;
        }
     case ispMoveETA:
        /* @q Move.ETA:Int (Ship Property)
           Estimated time of arrival at waypoint (number of turns). */
        if (sh.getShipKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(univ, sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getNumTurns());
        } else {
            return 0;
        }
     case ispMoveFuel:
        /* @q Move.Fuel:Int (Ship Property)
           Predicted fuel useage for movement, in kilotons. */
        if (sh.getShipKind() == sh.CurrentShip) {
            std::auto_ptr<GShipTurnPredictor> main, towee;
            if (sh.getMission() == GMission::msn_Tow && univ.ty_played_ships.isValidIndex(sh.getTowId())) {
                towee.reset(new GShipTurnPredictor(univ, sh.getTowId()));
                main.reset(new GShipTurnPredictor(univ, sh.getId(), *towee));
            } else {
                main.reset(new GShipTurnPredictor(univ, sh.getId()));
            }
            main->computeMovement();
            return makeIntValue(main->getMovementFuelUsed());
        } else {
            return 0;
        }
     case ispName:
        /* @q Name:Str (Ship Property)
           Ship name.
           @assignable
           @see SetName (Ship Command) */
        if (sh.isVisible())
            return makeStringValue(sh.getName(GObject::PlainName));
        else
            return 0;
     case ispOrbitId:
        /* @q Orbit$:Int (Ship Property)
           Id of planet this ship is orbiting. 0 if none. */
        if (sh.isVisible())
            return makeIntValue(sh.getOrbitPlanetId());
        else
            return 0;
     case ispOrbitName:
        /* @q Orbit:Str (Ship Property)
           Name of planet this ship is orbiting. EMPTY if none. */
        if (sh.getOrbitPlanetId() > 0)
            return makeStringValue(univ.getPlanet(sh.getOrbitPlanetId()).getName(GObject::PlainName));
        else
            return 0;
     case ispPlayed:
        /* @q Played:Bool (Ship Property)
           True if this ship is played.
           @since PCC 1.1.19 */
        return makeBoolValue(sh.isPlayable(GObject::Playable));
     case ispRealOwner:
        /* @q Owner.Real:Int (Ship Property)
           Real owner of this ship, player number.
           The real owner can differ from the {Owner (Ship Property)|Owner} reported normally
           when the ship is under remote control. */
        return makeIntValue(sh.getRealOwner());
     case ispSpeedId:
        /* @q Speed$:Int (Ship Property)
           Speed (warp factor).
           @assignable
           @see SetSpeed (Ship Command) */
        return makeIntFromMaybe(sh.getWarp());
     case ispSpeedName:
        /* @q Speed:Str (Ship Property)
           Speed, as human-readable string.
           If the hyperdrive is active, reports "Hyperdrive", otherwise "Warp x". */
        if (sh.getWarp().isKnown()) {
            if (sh.isHyperdriving())
                return makeStringValue("Hyperdrive");
            else
                return makeStringValue(format("Warp %d", sh.getWarp()));
        } else {
            return 0;
        }
     case ispTask:
        /* @q Task:Bool (Ship Property)
           True if this ship has an auto task. */
        return makeBoolValue(getProcessByObject(&sh, IntExecutionContext::pkShipTask) != 0);
     case ispTorpId:
        /* @q Torp$:Int (Ship Property, Combat Participant Property)
           Torpedo type. */
        return makeIntFromMaybe(sh.getTorpType());
     case ispTorpCount:
        /* @q Torp.Count:Int (Ship Property, Combat Participant Property)
           Number of torpedoes on this ship. 0 if the ship has no torpedoes. */
        if (sh.getTorpType().isKnown()) {
            if (sh.getTorpType() > 0)
                return makeIntFromMaybe(sh.getAmmo());
            else
                return makeIntValue(0);
        } else {
            return 0;
        }
     case ispTorpLCount:
        /* @q Torp.LCount:Int (Ship Property, Combat Participant Property)
           Number of torpedo launchers on this ship. */
        return makeIntFromMaybe(sh.getNumTorpLaunchers());
     case ispTorpShort:
        /* @q Torp.Short:Str (Ship Property, Combat Participant Property)
           Torpedo type, short name. */
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispTorpName:
        /* @q Torp:Str (Ship Property, Combat Participant Property)
           Torpedo type, full name. */
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispTransferShipColonists:
        /* @q Transfer.Ship.Colonists:Int (Ship Property)
           Number of colonists being transferred to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).colonists);
        else
            return 0;
     case ispTransferShipD:
        /* @q Transfer.Ship.D:Int (Ship Property)
           Amount of Duranium being transferred to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Duranium]);
        else
            return 0;
     case ispTransferShipId:
        /* @q Transfer.Ship.Id:Int (Ship Property)
           Id of cargo transfer target ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).id);
        else
            return 0;
     case ispTransferShipM:
        /* @q Transfer.Ship.M:Int (Ship Property)
           Amount of Molybdenum being transferred to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Molybdenum]);
        else
            return 0;
     case ispTransferShipN:
        /* @q Transfer.Ship.N:Int (Ship Property)
           Amount of Neutronium being transferred to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Neutronium]);
        else
            return 0;
     case ispTransferShipName:
        /* @q Transfer.Ship.Name:Str (Ship Property)
           Name of cargo transfer target ship. */
        if (sh.getShipKind() == sh.CurrentShip
            && sh.isTransporterActive(sh.t_Transfer)
            && univ.ty_any_ships.isValidIndex(sh.getTransporter(sh.t_Transfer).id))
        {
            return makeStringValue(univ.getShip(sh.getTransporter(sh.t_Transfer).id).getName(GObject::PlainName));
        } else {
            return 0;
        }
     case ispTransferShipT:
        /* @q Transfer.Ship.T:Int (Ship Property)
           Amount of Tritanium being transferred to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Tritanium]);
        else
            return 0;
     case ispTransferShipSupplies:
        /* @q Transfer.Ship.Supplies:Int (Ship Property)
           Amount of Supplies being transferred to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).supplies);
        else
            return 0;
     case ispTransferShip:
        /* @q Transfer.Ship:Bool (Ship Property)
           True if cargo is being transported to another ship. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Transfer));
        else
            return 0;
     case ispTransferUnloadColonists:
        /* @q Transfer.Unload.Colonists:Int (Ship Property)
           Number of colonists being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).colonists);
        else
            return 0;
     case ispTransferUnloadD:
        /* @q Transfer.Unload.D:Int (Ship Property)
           Amount of Duranium being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Duranium]);
        else
            return 0;
     case ispTransferUnloadId:
        /* @q Transfer.Unload.Id:Int (Ship Property)
           Id of planet cargo is being unloaded to. 0 for jettison. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).id);
        else
            return 0;
     case ispTransferUnloadM:
        /* @q Transfer.Unload.M:Int (Ship Property)
           Amount of Molybdenum being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Molybdenum]);
        else
            return 0;
     case ispTransferUnloadN:
        /* @q Transfer.Unload.N:Int (Ship Property)
           Amount of Neutronium being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Neutronium]);
        else
            return 0;
     case ispTransferUnloadName:
        /* @q Transfer.Unload.Name:Int (Ship Property)
           Name of planet cargo is being unloaded to. "Jettison" for jettison. */
        if (sh.getShipKind() == sh.CurrentShip && sh.isTransporterActive(sh.t_Unload)) {
            if (sh.getTransporter(sh.t_Unload).id == 0)
                return makeStringValue("Jettison");
            else if (univ.isValidPlanetId(sh.getTransporter(sh.t_Unload).id))
                return makeStringValue(univ.getPlanet(sh.getTransporter(sh.t_Unload).id).getName(GObject::PlainName));
            else
                return 0;
        } else {
            return 0;
        }
     case ispTransferUnloadT:
        /* @q Transfer.Unload.T:Int (Ship Property)
           Amount of Tritanium being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Tritanium]);
        else
            return 0;
     case ispTransferUnloadSupplies:
        /* @q Transfer.Unload.Supplies:Int (Ship Property)
           Amount of Supplies being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).supplies);
        else
            return 0;
     case ispTransferUnload:
        /* @q Transfer.Unload:Bool (Ship Property)
           True if cargo is being unloaded to a planet or deep space. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Unload));
        else
            return 0;
     case ispTypeChar:
        /* @q Type.Short:Str (Ship Property)
           Classification of ship, short.
           This is the first letter of the {Type (Ship Property)|Type}, see there. */
        if (const char* p = classifyShip(sh))
            return makeStringValue(string_t(p, 1));
        else
            return 0;
     case ispTypeStr:
        /* @q Type:Str (Ship Property)
           Classification of ship. Possible values are:
           - "Carrier"
           - "Torpedo Ship"
           - "Beam Weapons"
           - "Freighter" */
        if (const char* p = classifyShip(sh))
            return makeStringValue(p);
        else
            return 0;
     case ispWaypointDistance:
        /* @q Waypoint.Dist:Num (Ship Property)
           Distance to waypoint, in ly.
           This can be a fractional number. */
        if (sh.getShipKind() == sh.CurrentShip)
            return makeFloatValue(std::sqrt(double(square(sh.getWaypointDX()) + square(sh.getWaypointDY()))));
        else
            return 0;
     case ispWaypointDX:
        /* @q Waypoint.DX:Int (Ship Property)
           X distance to waypoint. */
        return makeIntFromMaybe(sh.getWaypointDX());
     case ispWaypointDY:
        /* @q Waypoint.DY:Int (Ship Property)
           Y distance to waypoint. */
        return makeIntFromMaybe(sh.getWaypointDY());
     case ispWaypointPlanetId:
        /* @q Waypoint.Planet:Int (Ship Property)
           Id of planet at waypoint.
           @see PlanetAt() */
        if (sh.getShipKind() == sh.CurrentShip) {
            return makeIntValue(univ.getPlanetAt(sh.getWaypoint(), false));
        } else {
            return 0;
        }
     case ispWaypointX:
        /* @q Waypoint.X:Int (Ship Property)
           X location of waypoint. */
        return makeIntFromMaybe(mp16_t(sh.getWaypoint().x));
     case ispWaypointY:
        /* @q Waypoint.Y:Int (Ship Property)
           Y location of waypoint. */
        return makeIntFromMaybe(mp16_t(sh.getWaypoint().y));
     case ispWaypointName:
        /* @q Waypoint:Str (Ship Property)
           Waypoint, as a human-readable string.
           @diff For ships on intercept course, PCC2 version 2.0.7 and later as well as PCC 1.x will return
           the name of the intercept target. Earlier PCC2 versions only return the name of the location
           (planet or X,Y's). */
        if (sh.getShipKind() == sh.CurrentShip) {
            int intId = sh.getInterceptId();
            if (sh.getMission() == GMission::msn_Intercept && univ.isValidShipId(intId)) {
                string_t name = univ.getShip(intId).getName(GObject::PlainName);
                if (name.empty()) {
                    return makeStringValue(format("Ship %d", intId));
                } else {
                    return makeStringValue(format("%s (#%d)", name, intId));
                }
            } else if (sh.getWaypointDX() == 0 && sh.getWaypointDY() == 0) {
                return makeStringValue("(Location)");
            } else {
                return makeStringValue(univ.getLocationName(sh.getWaypoint().getCanonicalLocation(), 0));
            }
        } else {
            return 0;
        }

     case ispScore:
     case ispHasFunction:
        return new IntShipArrayProperty(sh, isp);

    }
    return 0;
}

/** Modify ship property.
    \param sh Ship to modify
    \param isp Property to modify
    \param value New value */
void
setShipProperty(GShip& sh, GUniverse& univ, const IntShipProperty isp, IntValue* value)
{
    string_t sv;
    int32_t iv;

    // Everything is only assignable for own ships. As an exception, the name is also assignable for targets.
    if (sh.getShipKind() != GShip::CurrentShip) {
        if (isp == ispName && sh.getShipKind() != GShip::NoShip) {
            // accept
        } else {
            throw IntError::notAssignable();
        }
    }

    switch (isp) {
     case ispFCode:
        if (checkStringArg(sv, value))
            sh.setFCode(sv);
        break;
     case ispMissionId:
     case ispMissionIntercept:
     case ispMissionTow:
        if (checkIntArg(iv, value, 0, 10000)) {
            // Only CurrentShips for now, so no need to deal with unknowns
            int m = (isp == ispMissionId        ? iv : sh.getMission());
            int i = (isp == ispMissionIntercept ? iv : sh.getInterceptId());
            int t = (isp == ispMissionTow       ? iv : sh.getTowId());
            if (!setFleetMission(univ, sh.getId(), m, i, t)) {
                throw GError(GError::eFleet, GError::eFleet);
            }
        }
        break;
     case ispName:
        if (checkStringArg(sv, value))
            sh.setName(sv);
        break;
     case ispSpeedId:
        if (checkIntArg(iv, value, 0, 9)) {
            if (!setFleetSpeed(univ, sh.getId(), iv)) {
                throw GError(GError::eFleet, GError::eFleet);
            }
        }
        break;
     case ispEnemyId:
        if (checkIntArg(iv, value, 0, NUM_PLAYERS))
            sh.setPrimaryEnemy(mp16_t(iv));
        break;
     case ispFleetName:
        if (checkStringArg(sv, value)) {
            if (!setFleetName(univ, sh.getId(), sv)) {
                throw IntError::notAssignable();
            }
        }
        break;
     case ispFleetId:
        if (checkIntArg(iv, value, 0, NUM_SHIPS)) {
            if (!setFleetNumber(univ, sh.getId(), iv)) {
                throw IntError::rangeError();
            }
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

void
callShipMethod(GShip& sh, GUniverse& univ, IntShipMethod ism, IntExecutionContext& exc, IntArgBlock& args)
{
    ship_methods[ism](sh, univ, exc, args);
}
