/**
  *  \file u/t_int_parseexpr.cc
  *  \brief Test Expression Parsing and Evaluation
  *
  *  This module assumes absence of constant folding. With constant
  *  folding, the checkExecFailure tests might turn into compilation
  *  failures. A possible workaround would be to place the literals in
  *  global variables.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/parseexpr.h"

#include "u/t_int.h"

#define gva globalVars[0]
#define gvb globalVars[1]
#define gvc globalVars[2]

void
IntParseExprTestSuite::setUp()
{
    gva = gvb = gvc = 0;
}

void
IntParseExprTestSuite::tearDown()
{ }

/** Test all sorts of literals.
    Also tests the builtin 'Z' function which is used later on.
    Instructions: pushint, pushbool, pushlit, (pushe) */
void
IntParseExprTestSuite::testLiterals()
{
    // Literals
    checkBoolExpr("true", 1);
    checkBoolExpr("false", 0);
    checkIntExpr("1", 1);
    checkIntExpr("0", 0);
    checkIntExpr("99999", 99999);
    checkStringExpr("''", "");
    checkStringExpr("'foo'", "foo");
    checkStringExpr("\"\"", "");
    checkStringExpr("\"bar\"", "bar");
    checkFloatExpr("pi", 3.14159265);
    checkNullExpr("z(0)");
    checkNullExpr("#z(0)");
    checkNullExpr("# # # z(0)");
    checkExecFailure("#'foo'");
    checkExecFailure("#2+3");        // means plus(file(2),3) in PCC2
    checkFileExpr("#1", 1);
    checkFileExpr("#2", 2);
    checkFileExpr("#42", 42);
    checkFileExpr("# # # # 23", 23);
}

/** Test sequence ";" operator. */
void
IntParseExprTestSuite::testSequence()
{
    // operator ";"
    checkIntExpr("1;2", 2);
    checkIntExpr("1;2;3;4;5", 5);
    checkIntExpr("(1;2);3", 3);
    checkIntExpr("1;(3;4)", 4);
}

/** Test assignment ":=" operator. */
void
IntParseExprTestSuite::testAssignment()
{
    // operator ":="
    checkIntExpr("a:=1", 1);
    TS_ASSERT_EQUALS(gva, 1);
    gva = gvb = gvc = 0;

    checkIntExpr("a:=b:=1", 1);
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 1);
    gva = gvb = gvc = 0;

    checkIntExpr("(((a))):=17", 17);
    TS_ASSERT_EQUALS(gva, 17);
    gva = gvb = gvc = 0;
}

/** Test logical "Or" operator.
    Also uses ":=" and parentheses. */
void
IntParseExprTestSuite::testOr()
{
    // Test ternary logic
    checkBoolExpr("0 or 0",       false);
    checkBoolExpr("0 or 1",       true);
    checkBoolExpr("0 or 17",      true);
    checkNullExpr("0 or z(0)");
    checkBoolExpr("1 or 0",       true);
    checkBoolExpr("1 or 1",       true);
    checkBoolExpr("1 or z(0)",    true);
    checkBoolExpr("17 or 0",      true);
    checkBoolExpr("17 or 1",      true);
    checkBoolExpr("17 or z(0)",   true);
    checkNullExpr("z(0) or 0");
    checkBoolExpr("z(0) or 19",   true);
    checkNullExpr("z(0) or z(0)");

    // Check lazy evaluation
    checkBoolExpr("(a:=1; 0) or (b:=1; 0)", false);
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 1);

    checkBoolExpr("(a:=2; 0) or (b:=2; 1)", true);
    TS_ASSERT_EQUALS(gva, 2);
    TS_ASSERT_EQUALS(gvb, 2);

    checkBoolExpr("(a:=3; 1) or (b:=3; 0)", true);
    TS_ASSERT_EQUALS(gva, 3);
    TS_ASSERT_EQUALS(gvb, 2);

    checkNullExpr("(a:=4; z(0)) or (b:=4; 0)");
    TS_ASSERT_EQUALS(gva, 4);
    TS_ASSERT_EQUALS(gvb, 4);

    // Test 'if'
    checkIntExpr("if((a:=5; 0) or (b:=5; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 5);
    TS_ASSERT_EQUALS(gvb, 5);
    checkIntExpr("if((a:=6; 0) or (b:=6; 1),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 6);
    TS_ASSERT_EQUALS(gvb, 6);
    checkIntExpr("if((a:=7; 0) or (b:=7; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 7);
    TS_ASSERT_EQUALS(gvb, 7);

    checkIntExpr("if((a:=8; 1) or (b:=8; 0),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 8);
    TS_ASSERT_EQUALS(gvb, 7);
    checkIntExpr("if((a:=9; 1) or (b:=9; 1),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 9);
    TS_ASSERT_EQUALS(gvb, 7);
    checkIntExpr("if((a:=10; 1) or (b:=10; z(0)), 333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 7);

    checkIntExpr("if((a:=11; z(0)) or (b:=11; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 11);
    TS_ASSERT_EQUALS(gvb, 11);
    checkIntExpr("if((a:=12; z(0)) or (b:=12; 1),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 12);
    TS_ASSERT_EQUALS(gvb, 12);
    checkIntExpr("if((a:=13; z(0)) or (b:=13; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 13);
    TS_ASSERT_EQUALS(gvb, 13);
}

/** Test logical "And" operator.
    Also uses ":=" and parentheses. */
void
IntParseExprTestSuite::testAnd()
{
    // Test ternary logic
    checkBoolExpr("0 and 0",       false);
    checkBoolExpr("0 and 1",       false);
    checkBoolExpr("0 and 17",      false);
    checkBoolExpr("0 and z(0)",    false);
    checkBoolExpr("1 and 0",       false);
    checkBoolExpr("1 and 1",       true);
    checkNullExpr("1 and z(0)");
    checkBoolExpr("17 and 0",      false);
    checkBoolExpr("17 and 1",      true);
    checkNullExpr("17 and z(0)");
    checkBoolExpr("z(0) and 0",    false);
    checkNullExpr("z(0) and 19");
    checkNullExpr("z(0) and z(0)");

    // Check lazy evaluation
    checkBoolExpr("(a:=1; 0) and (b:=1; 0)", false);
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 0);

    checkBoolExpr("(a:=2; 0) and (b:=2; 1)", false);
    TS_ASSERT_EQUALS(gva, 2);
    TS_ASSERT_EQUALS(gvb, 0);

    checkBoolExpr("(a:=3; 1) and (b:=3; 0)", false);
    TS_ASSERT_EQUALS(gva, 3);
    TS_ASSERT_EQUALS(gvb, 3);

    checkBoolExpr("(a:=4; z(0)) and (b:=4; 0)", false);
    TS_ASSERT_EQUALS(gva, 4);
    TS_ASSERT_EQUALS(gvb, 4);

    checkNullExpr("(a:=5; z(0)) and (b:=5; 77)");
    TS_ASSERT_EQUALS(gva, 5);
    TS_ASSERT_EQUALS(gvb, 5);

    // Test 'if'
    gva = gvb = 0;
    checkIntExpr("if((a:=5; 0) and (b:=5; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 5);
    TS_ASSERT_EQUALS(gvb, 0);
    checkIntExpr("if((a:=6; 0) and (b:=6; 1),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 6);
    TS_ASSERT_EQUALS(gvb, 0);
    checkIntExpr("if((a:=7; 0) and (b:=7; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 7);
    TS_ASSERT_EQUALS(gvb, 0);

    checkIntExpr("if((a:=8; 1) and (b:=8; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 8);
    TS_ASSERT_EQUALS(gvb, 8);
    checkIntExpr("if((a:=9; 1) and (b:=9; 1),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 9);
    TS_ASSERT_EQUALS(gvb, 9);
    checkIntExpr("if((a:=10; 1) and (b:=10; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 10);

    // NOTE: here, the second part is not evaluated whereas in the similar 'a:=4' test, it is.
    // The reason is that we don't need an exact value here, and don't care whether the result
    // is empty or false.
    checkIntExpr("if((a:=11; z(0)) and (b:=11; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 11);
    TS_ASSERT_EQUALS(gvb, 10);
    checkIntExpr("if((a:=12; z(0)) and (b:=12; 1),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 12);
    TS_ASSERT_EQUALS(gvb, 10);
    checkIntExpr("if((a:=13; z(0)) and (b:=13; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 13);
    TS_ASSERT_EQUALS(gvb, 10);
}

/** Test logical "Xor" operator.
    Also uses ":=" and parentheses. */
void
IntParseExprTestSuite::testXor()
{
    // Test ternary logic
    checkBoolExpr("0 xor 0",       false);
    checkBoolExpr("0 xor 1",       true);
    checkBoolExpr("0 xor 17",      true);
    checkNullExpr("0 xor z(0)");
    checkBoolExpr("1 xor 0",       true);
    checkBoolExpr("1 xor 1",       false);
    checkNullExpr("1 xor z(0)");
    checkBoolExpr("17 xor 0",      true);
    checkBoolExpr("17 xor 1",      false);
    checkNullExpr("17 xor z(0)");
    checkNullExpr("z(0) xor 0");
    checkNullExpr("z(0) xor 19");
    checkNullExpr("z(0) xor z(0)");

    // Check lazy evaluation
    checkBoolExpr("(a:=1; 0) xor (b:=1; 0)", false);
    TS_ASSERT_EQUALS(gva, 1);
    TS_ASSERT_EQUALS(gvb, 1);

    checkBoolExpr("(a:=2; 0) xor (b:=2; 1)", true);
    TS_ASSERT_EQUALS(gva, 2);
    TS_ASSERT_EQUALS(gvb, 2);

    checkBoolExpr("(a:=3; 1) xor (b:=3; 0)", true);
    TS_ASSERT_EQUALS(gva, 3);
    TS_ASSERT_EQUALS(gvb, 3);

    checkNullExpr("(a:=4; z(0)) xor (b:=4; 0)");
    TS_ASSERT_EQUALS(gva, 4);
    TS_ASSERT_EQUALS(gvb, 3);

    // Test 'if'
    checkIntExpr("if((a:=5; 0) xor (b:=5; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 5);
    TS_ASSERT_EQUALS(gvb, 5);
    checkIntExpr("if((a:=6; 0) xor (b:=6; 1),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 6);
    TS_ASSERT_EQUALS(gvb, 6);
    checkIntExpr("if((a:=7; 0) xor (b:=7; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 7);
    TS_ASSERT_EQUALS(gvb, 7);

    checkIntExpr("if((a:=8; 1) xor (b:=8; 0),    333, 444)", 333);
    TS_ASSERT_EQUALS(gva, 8);
    TS_ASSERT_EQUALS(gvb, 8);
    checkIntExpr("if((a:=9; 1) xor (b:=9; 1),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 9);
    TS_ASSERT_EQUALS(gvb, 9);
    checkIntExpr("if((a:=10; 1) xor (b:=10; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 10);
    TS_ASSERT_EQUALS(gvb, 10);

    checkIntExpr("if((a:=11; z(0)) xor (b:=11; 0),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 11);
    TS_ASSERT_EQUALS(gvb, 10);
    checkIntExpr("if((a:=12; z(0)) xor (b:=12; 1),    333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 12);
    TS_ASSERT_EQUALS(gvb, 10);
    checkIntExpr("if((a:=13; z(0)) xor (b:=13; z(0)), 333, 444)", 444);
    TS_ASSERT_EQUALS(gva, 13);
    TS_ASSERT_EQUALS(gvb, 10);
}

/** Test logical "Not" operator.
    Instructions: unot, ubool. */
void
IntParseExprTestSuite::testNot()
{
    // Not
    checkBoolExpr("not 1", false);
    checkBoolExpr("not 0", true);
    checkNullExpr("not z(0)");

    checkBoolExpr("not not 1", true);
    checkBoolExpr("not not 99", true);
    checkBoolExpr("not not 0", false);
    checkNullExpr("not not z(0)");

    checkBoolExpr("not not not 1",  false);
    checkBoolExpr("not not not 99", false);
    checkBoolExpr("not not not 0",  true);
    checkNullExpr("not not not z(0)");

    // Test 'if'
    checkIntExpr("if(not 1, 333, 444)", 444);
    checkIntExpr("if(not 0, 333, 444)", 333);
    checkIntExpr("if(not z(0), 333, 444)", 444);

    checkIntExpr("if(not not 1, 333, 444)", 333);
    checkIntExpr("if(not not 0, 333, 444)", 444);
    checkIntExpr("if(not not z(0), 333, 444)", 444);
}

/** Test comparison operators: "=", "<>", "<", ">", "<=", ">=".
    Also tests the StrCase function.
    Instructions: bcmpeq, bcmpne, bcmplt, bcmpgt, bcmple, bcmpge, and bcmpXX_nc" versions of those. */
void
IntParseExprTestSuite::testComparison()
{
    // "="
    checkBoolExpr("1=1", true);
    checkBoolExpr("1=2", false);
    checkBoolExpr("1=1.0", true);
    checkBoolExpr("1.1=1.0", false);
    checkBoolExpr("1.0=1", true);
    checkBoolExpr("1.0=1.0", true);
    checkBoolExpr("'a'='A'", true);
    checkBoolExpr("'A'='A'", true);
    checkBoolExpr("strcase('a'='A')", false);
    checkBoolExpr("strcase('A'='A')", true);
    checkBoolExpr("''=''", true);
    checkNullExpr("z(0)=1");
    checkNullExpr("1=z(0)");
    checkNullExpr("0=z(0)");
    checkNullExpr("z(0)=z(0)");
    checkExecFailure("1='a'");
    checkExecFailure("'a'=1");
    checkExecFailure("'a'=1.0");
    checkExecFailure("'a'=a");

    checkBoolExpr("StrCase(1=1)", true);
    checkBoolExpr("StrCase(1=2)", false);
    checkBoolExpr("StrCase(1=1.0)", true);
    checkBoolExpr("StrCase(1.1=1.0)", false);
    checkBoolExpr("StrCase(1.0=1)", true);
    checkBoolExpr("StrCase(1.0=1.0)", true);
    checkNullExpr("StrCase(0=z(0))");
    checkNullExpr("StrCase(z(0)=z(0))");
    checkExecFailure("StrCase(1='a')");
    checkExecFailure("StrCase('a'=1)");
    
    // "<>"
    checkBoolExpr("1<>1", false);
    checkBoolExpr("1<>2", true);
    checkBoolExpr("1<>1.0", false);
    checkBoolExpr("1.1<>1.0", true);
    checkBoolExpr("1.0<>1", false);
    checkBoolExpr("1.0<>1.0", false);
    checkBoolExpr("'a'<>'A'", false);
    checkBoolExpr("'A'<>'A'", false);
    checkBoolExpr("strcase('a'<>'A')", true);
    checkBoolExpr("strcase('A'<>'A')", false);
    checkBoolExpr("''<>''", false);
    checkNullExpr("z(0)<>1");
    checkNullExpr("1<>z(0)");
    checkNullExpr("0<>z(0)");
    checkNullExpr("z(0)<>z(0)");
    checkExecFailure("1<>'a'");
    checkExecFailure("'a'<>1");
    checkExecFailure("'a'<>1.0");
    checkExecFailure("'a'<>a");

    // "<"
    checkBoolExpr("1<1", false);
    checkBoolExpr("1<2", true);
    checkBoolExpr("1<1.0", false);
    checkBoolExpr("1.1<1.0", false);
    checkBoolExpr("1.0<1.1", true);
    checkBoolExpr("1.0<1", false);
    checkBoolExpr("1.0<1.0", false);
    checkBoolExpr("'a'<'A'", false);
    checkBoolExpr("'A'<'A'", false);
    checkBoolExpr("strcase('a'<'A')", false);
    checkBoolExpr("strcase('A'<'A')", false);
    checkBoolExpr("''<''", false);
    checkBoolExpr("'a'<'b'", true);
    checkBoolExpr("'a'<''", false);
    checkNullExpr("z(0)<1");
    checkNullExpr("1<z(0)");
    checkNullExpr("0<z(0)");
    checkNullExpr("z(0)<z(0)");
    checkExecFailure("1<'a'");
    checkExecFailure("'a'<1");
    checkExecFailure("'a'<1.0");
    checkExecFailure("'a'<a");

    // ">"
    checkBoolExpr("1>1", false);
    checkBoolExpr("1>2", false);
    checkBoolExpr("1>1.0", false);
    checkBoolExpr("1.1>1.0", true);
    checkBoolExpr("1.0>1.1", false);
    checkBoolExpr("1.0>1", false);
    checkBoolExpr("1.0>1.0", false);
    checkBoolExpr("'a'>'A'", false);
    checkBoolExpr("'A'>'A'", false);
    checkBoolExpr("strcase('a'>'A')", true);
    checkBoolExpr("strcase('A'>'A')", false);
    checkBoolExpr("''>''", false);
    checkBoolExpr("'a'>'b'", false);
    checkBoolExpr("'a'>''", true);
    checkNullExpr("z(0)>1");
    checkNullExpr("1>z(0)");
    checkNullExpr("0>z(0)");
    checkNullExpr("z(0)>z(0)");
    checkExecFailure("1>'a'");
    checkExecFailure("'a'>1");
    checkExecFailure("'a'>1.0");
    checkExecFailure("'a'>a");

    // "<="
    checkBoolExpr("1<=1", true);
    checkBoolExpr("1<=2", true);
    checkBoolExpr("1<=1.0", true);
    checkBoolExpr("1.1<=1.0", false);
    checkBoolExpr("1.0<=1.1", true);
    checkBoolExpr("1.0<=1", true);
    checkBoolExpr("1.0<=1.0", true);
    checkBoolExpr("'a'<='A'", true);
    checkBoolExpr("'A'<='A'", true);
    checkBoolExpr("strcase('a'<='A')", false);
    checkBoolExpr("strcase('A'<='A')", true);
    checkBoolExpr("''<=''", true);
    checkBoolExpr("'a'<='b'", true);
    checkBoolExpr("'a'<=''", false);
    checkNullExpr("z(0)<=1");
    checkNullExpr("1<=z(0)");
    checkNullExpr("0<=z(0)");
    checkNullExpr("z(0)<=z(0)");
    checkExecFailure("1<='a'");
    checkExecFailure("'a'<=1");
    checkExecFailure("'a'<=1.0");
    checkExecFailure("'a'<=a");

    // ">="
    checkBoolExpr("1>=1", true);
    checkBoolExpr("1>=2", false);
    checkBoolExpr("1>=1.0", true);
    checkBoolExpr("1.1>=1.0", true);
    checkBoolExpr("1.0>=1.1", false);
    checkBoolExpr("1.0>=1", true);
    checkBoolExpr("1.0>=1.0", true);
    checkBoolExpr("'a'>='A'", true);
    checkBoolExpr("'A'>='A'", true);
    checkBoolExpr("strcase('a'>='A')", true);
    checkBoolExpr("strcase('A'>='A')", true);
    checkBoolExpr("''>=''", true);
    checkBoolExpr("'a'>='b'", false);
    checkBoolExpr("'a'>=''", true);
    checkNullExpr("z(0)>=1");
    checkNullExpr("1>=z(0)");
    checkNullExpr("0>=z(0)");
    checkNullExpr("z(0)>=z(0)");
    checkExecFailure("1>='a'");
    checkExecFailure("'a'>=1");
    checkExecFailure("'a'>=1.0");
    checkExecFailure("'a'>=a");
}

/** Test concatenation operators: "&", "#".
    Instructions: bconcat, bconcatempty. */
void
IntParseExprTestSuite::testConcat()
{
    // "&", interpolates Empty as ''
    checkStringExpr("1 & 2", "12");
    checkStringExpr("1 & 'a'", "1a");
    checkStringExpr("1 & z(0)", "1");
    checkNullExpr("z(0) & z(0)");
    checkStringExpr("'a' & 'b' & 'c' & 'd' & z(0)", "abcd");
    checkStringExpr("'a' & 'b' & 'c' & z(0) & 'd'", "abcd");
    checkStringExpr("'a' & 'b' & z(0) & 'c' & 'd'", "abcd");
    checkStringExpr("'a' & z(0) & 'b' & 'c' & 'd'", "abcd");
    checkStringExpr("z(0) & 'a' & 'b' & 'c' & 'd'", "abcd");

    // "#", Empty annihilates expression
    checkStringExpr("1 # 2", "12");
    checkStringExpr("1 # 'a'", "1a");
    checkNullExpr("1 # z(0)");
    checkNullExpr("z(0) # z(0)");
    checkNullExpr("'a' # 'b' # 'c' # 'd' # z(0)");
    checkNullExpr("'a' # 'b' # 'c' # z(0) # 'd'");
    checkNullExpr("'a' # 'b' # z(0) # 'c' # 'd'");
    checkNullExpr("'a' # z(0) # 'b' # 'c' # 'd'");
    checkNullExpr("z(0) # 'a' # 'b' # 'c' # 'd'");
}

/** Test addition operator "+".
    Instructions: badd. */
void
IntParseExprTestSuite::testAdd()
{
    // Integers
    checkIntExpr("1 + 1", 2);
    checkIntExpr("0 + 1000000", 1000000);
    checkNullExpr("1 + z(0)");
    checkNullExpr("z(0) + 1");

    // Floats
    checkFloatExpr("1.0 + 2.0", 3.0);
    checkFloatExpr("0.0 + 1000000.0", 1000000.0);
    checkNullExpr("1.0 + z(0)");
    checkNullExpr("z(0) + 1.0");

    // Mixed
    checkFloatExpr("1 + 2.0", 3.0);
    checkFloatExpr("0.0 + 0", 0.0);

    // Strings
    checkStringExpr("'a' + 'b'", "ab");
    checkNullExpr("'a' + z(0)");
    checkNullExpr("z(0) + 'a'");

    // Errors
    checkExecFailure("'a' + 1");
    checkExecFailure("1 + 'a'");
}

/** Test subtraction operator: "-".
    Instruction: bsub. */
void
IntParseExprTestSuite::testSubtract()
{
    // Integers
    checkIntExpr("1 - 1", 0);
    checkIntExpr("100 - 1", 99);
    checkIntExpr("0 - 1000000", -1000000);
    checkNullExpr("1 - z(0)");
    checkNullExpr("z(0) - 1");

    // Floats
    checkFloatExpr("1.0 - 2.0", -1.0);
    checkFloatExpr("0.0 - 1000000.0", -1000000.0);
    checkNullExpr("1.0 - z(0)");
    checkNullExpr("z(0) - 1.0");

    // Mixed
    checkFloatExpr("12 - 2.0", 10.0);
    checkFloatExpr("0.0 - 0", 0.0);

    // Errors
    checkExecFailure("'a' - 1");
    checkExecFailure("'ab' - 'a'");
    // checkExecFailure("'a' - z(0)");  Not an error - should it?
    // checkExecFailure("z(0) - 'a'");  Not an error - should it?
}

/** Test multiplication operator: "*".
    Instruction: bmul. */
void
IntParseExprTestSuite::testMultiply()
{
    // Integers
    checkIntExpr("2*3*4", 24);
    checkIntExpr("10*0", 0);
    checkNullExpr("z(0) * 10");
    checkNullExpr("10 * z(0)");

    // Floats
    checkFloatExpr("2.0*3.0*4.0", 24.0);
    checkFloatExpr("10.0 * 0", 0.0);
    checkNullExpr("z(0) * 10.0");
    checkNullExpr("10.0 * z(0)");

    // Mixed
    checkFloatExpr("2*3.0", 6);
    checkFloatExpr("2.0*3", 6);

    // Errors
    checkExecFailure("10*'a'");
    checkExecFailure("'a'*10");
    // checkExecFailure("'a'*z(0)");  Not an error - should it?
    // checkExecFailure("z(0)*'a'");  Not an error - should it?
}

/** Test real division operator "/".
    Instruction: bdiv. */
void
IntParseExprTestSuite::testDivide()
{
    // Integers
    checkIntExpr("16/4", 4);
    checkIntExpr("10/1", 10);
    checkFloatExpr("5/2", 2.5);
    checkNullExpr("z(0) / 10");
    checkNullExpr("10 / z(0)");

    // Floats
    checkFloatExpr("16.0/4.0", 4.0);
    checkFloatExpr("2.0/4.0", 0.5);
    checkFloatExpr("10.0/4.0", 2.5);
    checkNullExpr("z(0) / 10.0");
    checkNullExpr("10.0 / z(0)");

    // Mixed
    checkFloatExpr("2 / 4.0", 0.5);
    checkFloatExpr("2.0 / 4", 0.5);

    // Errors
    checkExecFailure("10/'a'");
    checkExecFailure("'a'/10");
    // checkExecFailure("'a'/z(0)");  Not an error - should it?
    // checkExecFailure("z(0)/'a'");  Not an error - should it?

    checkExecFailure("10/0");
    checkExecFailure("10.0/0");
    checkExecFailure("10.0/0.0");
}

/** Test integral division operators: "\", "Mod".
    Instructions: bidiv, brem. */
void
IntParseExprTestSuite::testIntegerDivide()
{
    // Integers
    checkIntExpr("15 \\ 3", 5);
    checkIntExpr("16 \\ 3", 5);
    checkIntExpr("17 \\ 3", 5);
    checkIntExpr("18 \\ 3", 6);
    checkIntExpr("15 mod 3", 0);
    checkIntExpr("16 mod 3", 1);
    checkIntExpr("17 mod 3", 2);
    checkIntExpr("18 mod 3", 0);

    checkNullExpr("z(0) \\ 3");
    checkNullExpr("15 \\ z(0)");
    checkNullExpr("z(0) mod 3");
    checkNullExpr("15 mod z(0)");

    // Floats
    checkExecFailure("15.0 \\ 3");
    checkExecFailure("15 \\ 3.0");
    checkExecFailure("15.0 mod 3");
    checkExecFailure("15 mod 3.0");

    // Errors
    checkExecFailure("'a' \\ 3");
    checkExecFailure("3 \\ 'a'");
    checkExecFailure("'a' mod 3");
    checkExecFailure("3 mod 'a'");
    // checkExecFailure("'a' \\ z(0)");  Not an error - should it?
    // checkExecFailure("z(0) \\ 'a'");  Not an error - should it?
}

/** Test unary signs "+", "-".
    Instructions: uneg, upos. */
void
IntParseExprTestSuite::testNegation()
{
    // Integers
    checkIntExpr("-1", -1);
    checkIntExpr("+1", 1);

    checkIntExpr("--1", 1);
    checkIntExpr("+-1", -1);
    checkIntExpr("-+1", -1);
    checkIntExpr("++1", 1);

    checkIntExpr("---1", -1);
    checkIntExpr("+--1", 1);
    checkIntExpr("-+-1", 1);
    checkIntExpr("++-1", -1);
    checkIntExpr("--+1", 1);
    checkIntExpr("+-+1", -1);
    checkIntExpr("-++1", -1);
    checkIntExpr("+++1", 1);

    // Floats
    checkFloatExpr("-1.0", -1.0);
    checkFloatExpr("+1.0", 1.0);

    checkFloatExpr("--1.0", 1.0);
    checkFloatExpr("+-1.0", -1.0);
    checkFloatExpr("-+1.0", -1.0);
    checkFloatExpr("++1.0", 1.0);

    checkFloatExpr("---1.0", -1.0);
    checkFloatExpr("+--1.0", 1.0);
    checkFloatExpr("-+-1.0", 1.0);
    checkFloatExpr("++-1.0", -1.0);
    checkFloatExpr("--+1.0", 1.0);
    checkFloatExpr("+-+1.0", -1.0);
    checkFloatExpr("-++1.0", -1.0);
    checkFloatExpr("+++1.0", 1.0);

    // Strings
    checkExecFailure("+'a'");
    checkExecFailure("-'a'");
    checkExecFailure("+-'a'");
    checkExecFailure("--'a'");
    checkExecFailure("++'a'");
    checkExecFailure("-+'a'");

    // Null
    checkNullExpr("-z(0)");
    checkNullExpr("+z(0)");
    checkNullExpr("-+z(0)");
    checkNullExpr("++z(0)");
    checkNullExpr("+-z(0)");
    checkNullExpr("--z(0)");
}

/** Test exponentiation operator "^".
    Instruction: bpow. */
void
IntParseExprTestSuite::testPower()
{
    // Integers
    checkIntExpr("2^8", 256);
    checkIntExpr("0^10", 0);
    checkIntExpr("10^0", 1);
    checkIntExpr("61^2", 3721);
    checkIntExpr("-61^2", -3721);
    checkIntExpr("(-61)^2", 3721);

    // Boundaries
    checkIntExpr("46340^2", 2147395600);
    checkFloatExpr("46341^2", 2147488281.0);
    checkIntExpr("2^20", 1048576);
    checkFloatExpr("3^20", 3486784401.0);
    checkFloatExpr("3^31", 617673396283947.0);

    // Floats
    checkFloatExpr("10^12", 1000000000000.0);
    // checkFloatExpr("4^0.5", 2);
    checkFloatExpr("0.5^2", 0.25);
    // checkFloatExpr("4^2.5", 32);

    // Null
    checkNullExpr("2^z(0)");
    checkNullExpr("z(0)^2");
    // checkNullExpr("z(0)^2.5");
    checkNullExpr("z(0)^3");
    checkNullExpr("z(0)^z(0)");

    // Strings
    checkExecFailure("2^'a'");
    checkExecFailure("'a'^3");
    checkExecFailure("'a'^'b'");
    // checkExecFailure("'a' ^ z(0)");
    // checkExecFailure("z(0) ^ 'a'");

    // Parsing
    checkIntExpr("-3^2", -9);
    checkIntExpr("(-3)^2", 9);
    checkFloatExpr("3^-2", 0.1111111111111111111111111111111111111111111111111111);
}

void
IntParseExprTestSuite::testPrecedence()
{
    checkIntExpr("1+2*3", 7);
    checkIntExpr("1*2+3", 5);
    checkIntExpr("(1+2)*3", 9);
    checkIntExpr("1+2^3*4", 33);
    checkBoolExpr("1 or 1 and 0", true);
    checkBoolExpr("(1 or 1) and 0", false);
    checkBoolExpr("1 or (1 and 0)", true);
}
