/**
  *  \file ui/scroll.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/scroll.h"
#include "gfx/clip.h"

UIScrollableGroup::UIScrollableGroup(UIBaseWidget& content)
    : UIScrollable(0),
      content(content),
      scrollbar(*this)
{
    addChildWidget(content, 0);
}

UIScrollableGroup::~UIScrollableGroup()
{ }

void
UIScrollableGroup::onScroll()
{
    GfxRect r(content.getExtent());
    r.y = getExtent().y - getCurrent();
    content.setExtent(r, false);
    content.drawWidget(false);
}

void
UIScrollableGroup::onLayout(bool by_layout)
{
    LayoutInfo li;
    content.getLayoutInfo(li);
    setSize(li.pref_size.y);
    setPageSize(getExtent().y);
    adjustPosition();

    GfxRect me(getExtent());
    if (getSize() > getPageSize()) {
        /* We need a scrollbar */
        LayoutInfo lisb;
        scrollbar.getLayoutInfo(lisb);
        scrollbar.setExtent(GfxRect(me.x + me.w - lisb.pref_size.x,
                                    me.y,
                                    lisb.pref_size.x,
                                    me.h),
                            by_layout);
        content.setExtent(GfxRect(me.x,
                                  me.y - getCurrent(),
                                  me.w - lisb.pref_size.x,
                                  li.pref_size.y),
                          by_layout);
        if (scrollbar.getParent() == 0)
            addChildWidget(scrollbar, 0);
    } else {
        /* We don't need a scrollbar */
        content.setExtent(GfxRect(me.x,
                                  me.y - getCurrent(),
                                  me.w,
                                  li.pref_size.y),
                          by_layout);
        if (scrollbar.getParent() != 0)
            removeChildWidget(scrollbar);
    }
}

// Drawing
void
UIScrollableGroup::drawContent(GfxCanvas& /*can*/)
{
    // we have no content but our children
}

void
UIScrollableGroup::drawEverything(GfxCanvas& can)
{
    drawContent(can); /* for completeness */

    if (scrollbar.hasState(st_Exposed)) {
        /* Child widget and scrollbar */
        GfxRect r(getExtent());
        r.w -= scrollbar.getExtent().w;
        GfxClipRectFilter filter(r, can);
        content.drawEverything(filter);
        scrollbar.drawEverything(can);
    } else {
        /* Just single child widget */
        GfxClipRectFilter filter(getExtent(), can);
        content.drawEverything(filter);
    }
}

void
UIScrollableGroup::drawBelow(UIBaseWidget& /*child*/, GfxCanvas& can)
{
    if (getParent())
        getParent()->drawBelow(*this, can);
}

GfxCanvas&
UIScrollableGroup::getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr)
{
    if (&child == &scrollbar) {
        return getParent()->getCanvasFor(*this, mgr);
    } else {
        GfxRect r(getExtent());
        if (scrollbar.hasState(st_Exposed))
            r.w -= scrollbar.getExtent().w;
        return mgr.add(new GfxClipRectFilter(getExtent(), getParent()->getCanvasFor(*this, mgr)));
    }
}

// Resizing
void
UIScrollableGroup::onResize(GfxRect /*oldsize*/, bool by_layout)
{
    onLayout(by_layout);
}

void
UIScrollableGroup::onChildResize(UIBaseWidget& /*child*/, GfxRect /*oldsize*/)
{
    // onLayout(false);
}

void
UIScrollableGroup::onAddRemoveChild(UIBaseWidget& /*the_widget*/, bool /*added*/)
{
    // onLayout(false);
}

// Layout
void
UIScrollableGroup::getLayoutInfo(LayoutInfo& info)
{
    scrollbar.getLayoutInfo(info);
    info.pref_size = info.min_size = info.pref_size.scaledBy(2, 2);
    info.flags     = li_Flexible;
}

void
UIScrollableGroup::scrollTo(int pos, int /*corr*/)
{
    if (pos < 0)
        pos = 0;
    if (pos >= getSize())
        pos = getSize()-1;
    if (pos != getCurrent()) {
        setCurrent(pos);
        adjustPosition();
        onScroll();
    }
}
