/**
  *  \file util/helpindex.h
  *  \brief Class HelpIndex
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_HELPINDEX_H
#define PCC_V2_UTIL_HELPINDEX_H

#include <vector>
#include <map>
#include "io/streams.h"
#include "cpluslib/ptrvector.h"

/** Help index.
    Manages a set of help files.

    Help files are XML files consisting of pages
       <page id="name:of:the:page">...</page>
    Pages can be grouped and given a priority
       <group priority=123><page...></page>...</group>
       <help priority=123><page...></page>...</group>

    This class pre-parses the help files to find all pages and associated priorities.
    It does not interpret the content of the pages, it just needs to be balanced XML.

    Users can then obtain a list of file positions to combine into a help page:
    - if multiple pages are provided with different priorities, those are concatenated in increasing priority order.
      Pages that have no priority is given are treated as priority 100.
    - if multiple pages are provided with the same priority, the one from the file added last is used
      (this is the case that the help file is replaced (overlaid) by a different file). */
class HelpIndex {
 public:
    struct File {
        string_t name;
        string_t origin;
        bool scanned;
        int serial;
    };
    struct Node {
        int priority;
        const File& file;
        Stream::fpos_t pos;
        Node(int pri, const File& file, Stream::fpos_t pos)
            : priority(pri), file(file), pos(pos)
            { }
    };

    HelpIndex();
    ~HelpIndex();

    void addFile(string_t name, string_t origin);
    void removeFilesByOrigin(string_t origin);

    void find(string_t page, std::vector<Node*>& out);

    static HelpIndex instance;

 private:
    struct Page {
        std::vector<Node> nodes;
    };
    ptr_vector<File> m_files;
    std::multimap<string_t, Node> m_nodes;
    int m_counter;

    void scanNewFiles();
    void scanFile(const File& file);
};

#endif
